import functools
import operator
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

import pandas
import pendulum

from tecton_core import errors
from tecton_core import time_utils
from tecton_core.id_helper import IdHelper
from tecton_proto.args.pipeline__client_pb2 import ConstantNode
from tecton_proto.args.pipeline__client_pb2 import DataSourceNode
from tecton_proto.args.pipeline__client_pb2 import Input as InputProto
from tecton_proto.args.pipeline__client_pb2 import Pipeline
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.args.pipeline__client_pb2 import RequestContext as RequestContextProto
from tecton_proto.args.pipeline__client_pb2 import RequestDataSourceNode
from tecton_proto.args.pipeline__client_pb2 import TransformationNode


CONSTANT_TYPE = Optional[Union[str, int, float, bool]]
CONSTANT_TYPE_OBJECTS = (str, int, float, bool)


def _make_mode_to_type() -> Dict[str, Any]:
    lookup: Dict[str, Any] = {
        "pandas": pandas.DataFrame,
        "python": Dict,
        "pipeline": PipelineNode,
        "spark_sql": str,
        "snowflake_sql": str,
    }
    try:
        # The `pyspark.sql.connect.*` classes are the default for the Notebooks on Databricks Shared Access Mode
        # Clusters of Version 14.3 or higher. That's why we need to support them.
        from pyspark.sql import DataFrame

        try:
            from pyspark.sql.connect.session import DataFrame as ConnectDataFrame
        except ImportError:
            ConnectDataFrame = DataFrame

        lookup["pyspark"] = (DataFrame, ConnectDataFrame)
    except ImportError:
        pass
    try:
        import snowflake.snowpark

        lookup["snowpark"] = snowflake.snowpark.DataFrame
    except ImportError:
        pass
    return lookup


MODE_TO_TYPE_LOOKUP: Dict[str, Any] = _make_mode_to_type()


def constant_node_to_value(constant_node: ConstantNode) -> CONSTANT_TYPE:
    if constant_node.HasField("string_const"):
        return constant_node.string_const
    elif constant_node.HasField("int_const"):
        return int(constant_node.int_const)
    elif constant_node.HasField("float_const"):
        return float(constant_node.float_const)
    elif constant_node.HasField("bool_const"):
        return constant_node.bool_const
    elif constant_node.HasField("null_const"):
        return None
    msg = f"Unknown ConstantNode type: {constant_node}"
    raise KeyError(msg)


def get_keyword_inputs(transformation_node: TransformationNode) -> Dict[str, InputProto]:
    """Returns the keyword inputs of transformation_node in a dict."""
    return {
        node_input.arg_name: node_input for node_input in transformation_node.inputs if node_input.HasField("arg_name")
    }


def positional_inputs(transformation_node: TransformationNode) -> List[InputProto]:
    """Returns the positional inputs of transformation_node in order."""
    return [node_input for node_input in transformation_node.inputs if node_input.HasField("arg_index")]


def check_transformation_type(
    object_name: str,
    result: Any,  # noqa: ANN401
    mode: str,
    supported_modes: List[str],
) -> None:
    possible_mode = None
    for candidate_mode, candidate_type in MODE_TO_TYPE_LOOKUP.items():
        if isinstance(result, candidate_type):
            possible_mode = candidate_mode
            break
    expected_type = MODE_TO_TYPE_LOOKUP[mode]
    actual_type = type(result)

    if isinstance(result, expected_type):
        return
    elif possible_mode is not None and possible_mode in supported_modes:
        msg = f"Transformation function {object_name} with mode '{mode}' is expected to return result with type {expected_type}, but returns result with type {actual_type} instead. Did you mean to set mode='{possible_mode}'?"
        raise TypeError(msg)
    else:
        msg = f"Transformation function {object_name} with mode {mode} is expected to return result with type {expected_type}, but returns result with type {actual_type} instead."
        raise TypeError(msg)


def get_time_window_from_data_source_node(
    feature_time_limits: Optional[pendulum.Period],
    schedule_interval: Optional[pendulum.Duration],
    data_source_node: DataSourceNode,
) -> Optional[pendulum.Period]:
    if data_source_node.HasField("window") and feature_time_limits:
        new_start = feature_time_limits.start - time_utils.proto_to_duration(data_source_node.window)
        if schedule_interval:
            new_start = new_start + schedule_interval
        raw_data_limits = pendulum.Period(new_start, feature_time_limits.end)
    elif data_source_node.HasField("window_unbounded_preceding") and feature_time_limits:
        raw_data_limits = pendulum.Period(pendulum.datetime(1970, 1, 1), feature_time_limits.end)
    elif data_source_node.HasField("start_time_offset") and feature_time_limits:
        new_start = feature_time_limits.start + time_utils.proto_to_duration(data_source_node.start_time_offset)
        raw_data_limits = pendulum.Period(new_start, feature_time_limits.end)
    elif data_source_node.HasField("window_unbounded"):
        raw_data_limits = None
    else:
        # no data_source_override has been set
        raw_data_limits = feature_time_limits
    return raw_data_limits


# TODO(jiadong): Consolidate this method with `get_request_context_node` as they share similar functionality. Need to migrate off all usages of this method first.
def find_request_context(node: PipelineNode) -> Optional[RequestContextProto]:
    """Returns the request context for the pipeline. Assumes there is at most one RequestContext."""
    if node.HasField("request_data_source_node"):
        return node.request_data_source_node.request_context
    elif node.HasField("transformation_node"):
        for child in node.transformation_node.inputs:
            rc = find_request_context(child.node)
            if rc is not None:
                return rc
    return None


def get_input_name_to_ds_id_map(pipeline: Pipeline) -> Dict[str, str]:
    """Return a map from input name to data source id for the pipeline."""
    data_source_nodes = get_all_data_source_nodes(pipeline)
    return {
        node.data_source_node.input_name: IdHelper.to_string(node.data_source_node.virtual_data_source_id)
        for node in data_source_nodes
    }


def get_request_context_node(pipeline: Pipeline) -> Optional[RequestDataSourceNode]:
    """Returns the request_data_source_node for the pipeline. Assumes there is at most one RequestContext."""
    rc_node = [node for node in get_all_pipeline_nodes(pipeline.root) if node.HasField("request_data_source_node")]

    if len(rc_node) == 0:
        return None
    elif len(rc_node) == 1:
        return rc_node[0].request_data_source_node
    else:
        msg = "ODFV is not supposed to have more than 1 request_data_source_node"
        raise errors.TectonValidationError(msg)


def get_all_feature_view_nodes(pipeline: Pipeline) -> List[PipelineNode]:
    """Returns all feature view nodes from the provided pipeline."""
    return [node for node in get_all_pipeline_nodes(pipeline.root) if node.HasField("feature_view_node")]


def get_all_data_source_nodes(pipeline: Pipeline) -> List[PipelineNode]:
    """Returns all data source nodes from the provided pipeline."""
    return [node for node in get_all_pipeline_nodes(pipeline.root) if node.HasField("data_source_node")]


def get_all_pipeline_nodes(node: PipelineNode) -> List[PipelineNode]:
    """Returns all data source nodes from the provided node."""
    if node.HasField("transformation_node"):
        return functools.reduce(
            operator.iadd, [get_all_pipeline_nodes(input.node) for input in node.transformation_node.inputs], []
        )
    else:
        return [node]


def get_all_input_keys(node: PipelineNode) -> Set[str]:
    names_set = set()
    _get_all_input_keys_helper(node, names_set)
    return names_set


def _get_all_input_keys_helper(node: PipelineNode, names_set: Set[str]) -> Set[str]:
    if node.HasField("request_data_source_node"):
        names_set.add(node.request_data_source_node.input_name)
    elif node.HasField("data_source_node"):
        names_set.add(node.data_source_node.input_name)
    elif node.HasField("feature_view_node"):
        names_set.add(node.feature_view_node.input_name)
    elif node.HasField("context_node"):
        names_set.add(node.context_node.input_name)
    elif node.HasField("transformation_node"):
        for child in node.transformation_node.inputs:
            _get_all_input_keys_helper(child.node, names_set)
    return names_set


def get_fco_ids_to_input_keys(node: PipelineNode) -> Dict[str, str]:
    names_dict = {}
    _get_fco_ids_to_input_keys_helper(node, names_dict)
    return names_dict


def _get_fco_ids_to_input_keys_helper(node: PipelineNode, names_dict: Dict[str, str]) -> Dict[str, str]:
    if node.HasField("request_data_source_node"):
        # request data sources don't have fco ids
        pass
    elif node.HasField("data_source_node"):
        ds_node = node.data_source_node
        names_dict[IdHelper.to_string(ds_node.virtual_data_source_id)] = ds_node.input_name
    elif node.HasField("feature_view_node"):
        fv_node = node.feature_view_node
        names_dict[IdHelper.to_string(fv_node.feature_view_id)] = fv_node.input_name
    elif node.HasField("transformation_node"):
        for child in node.transformation_node.inputs:
            _get_fco_ids_to_input_keys_helper(child.node, names_dict)
    return names_dict


def uses_realtime_context(node: PipelineNode) -> bool:
    """Returns true if a Transformation with a RealtimeContext input exists in the Pipeline"""
    if node.HasField("context_node"):
        return True
    elif node.HasField("transformation_node"):
        return any(uses_realtime_context(child.node) for child in node.transformation_node.inputs)



import copy
import enum
from types import MappingProxyType
from typing import Any
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pendulum
from google.protobuf import duration_pb2
from typeguard import typechecked

from tecton_core import conf
from tecton_core import data_types as tecton_types
from tecton_core import errors
from tecton_core import feature_view_utils
from tecton_core import id_helper
from tecton_core import pipeline_common
from tecton_core import schema
from tecton_core import time_utils
from tecton_core.aggregation_utils import get_aggregation_enum_from_string
from tecton_core.compute_mode import BatchComputeMode
from tecton_core.specs import tecton_object_spec
from tecton_core.specs import utils
from tecton_core.specs.time_window_spec import LifetimeWindowSpec
from tecton_core.specs.time_window_spec import RelativeTimeWindowSpec
from tecton_core.specs.time_window_spec import TimeWindowSeriesSpec
from tecton_core.specs.time_window_spec import TimeWindowSpec
from tecton_core.specs.time_window_spec import create_time_window_spec_from_data_proto
from tecton_core.specs.time_window_spec import window_spec_to_window_data_proto
from tecton_proto.args import feature_view__client_pb2 as feature_view__args_pb2
from tecton_proto.args import pipeline__client_pb2 as pipeline_pb2
from tecton_proto.common import aggregation_function__client_pb2 as afpb
from tecton_proto.common import data_source_type__client_pb2 as data_source_type_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_proto.data import feature_store__client_pb2 as feature_store_pb2
from tecton_proto.data import feature_view__client_pb2 as feature_view__data_pb2
from tecton_proto.modelartifactservice import model_artifact_service__client_pb2 as model_artifact_service_pb2
from tecton_proto.validation import validator__client_pb2 as validator_pb2


__all__ = [
    "FeatureViewSpec",
    "MaterializedFeatureViewSpec",
    "RealtimeFeatureViewSpec",
    "FeatureTableSpec",
    "MaterializedFeatureViewType",
    "OnlineBatchTablePart",
    "OnlineBatchTablePartTile",
    "create_feature_view_spec_from_data_proto",
    "create_feature_view_spec_from_args_proto",
    "FeatureViewSpecArgsSupplement",
    "get_batch_schedule_from_feature_view_args",
    "get_batch_trigger_from_feature_view_args",
    "get_is_continuous_from_feature_view_args",
    "get_join_keys_from_feature_view_args",
    "get_online_serving_keys_from_feature_view_args",
    "get_aggregate_features_from_feature_view_args",
    "resolve_timestamp_field",
]


@utils.frozen_strict
class FeatureViewSpec(tecton_object_spec.TectonObjectSpec):
    """Base class for feature view specs."""

    join_keys: Tuple[str, ...]
    entity_ids: Tuple[str, ...]
    online_serving_keys: Tuple[str, ...]  # Aka the Online Serving Index.
    feature_store_format_version: feature_store_pb2.FeatureStoreFormatVersion.ValueType = attrs.field()
    view_schema: schema.Schema
    materialization_schema: schema.Schema

    # materialization_enabled is True if the feature view has online or online set to True, and the feature view is
    # applied to a live workspace.
    materialization_enabled: bool
    online: bool
    offline: bool
    options: Mapping[str, str]

    url: Optional[str] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})
    context_parameter_name: Optional[str]

    @feature_store_format_version.validator
    def check_valid_feature_store_format_version(self, _, value):
        if (
            value < feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT
            or value > feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_MAX
        ):
            msg = f"Unsupported feature_store_format_version: {value}"
            raise ValueError(msg)

    @property
    def features(self) -> List[str]:
        """
        Returns the output feature names of this feature view
        """
        raise NotImplementedError

    def join_key_schema(self) -> schema.Schema:
        """
        Return the join key schmea which is adjusted by online serving keys. Only ODFV needs to overwrite this.
        """
        spine_schema_dict = self.view_schema.to_dict()
        # If online_serving_keys are specified, spine only needs to contain these keys instead of all join keys.
        retrieval_keys = self.online_serving_keys if len(self.online_serving_keys) > 0 else self.join_keys
        spine_schema_dict = {key: spine_schema_dict[key] for key in retrieval_keys}
        return schema.Schema.from_dict(spine_schema_dict)


@attrs.define
class FeatureViewSpecArgsSupplement:
    """A data class used for supplementing args protos during FeatureViewSpec construction.

    This Python data class can be used to include data that is not included in args protos (e.g. schemas) into the
    FeatureViewSpec constructor.
    """

    view_schema: Optional[schema_pb2.Schema]
    materialization_schema: Optional[schema_pb2.Schema]
    online_batch_table_format: Optional[schema_pb2.OnlineBatchTableFormat]


@utils.frozen_strict
class SecondaryKeyOutputColumn:
    name: str
    time_window: TimeWindowSpec


@utils.frozen_strict
class OnlineBatchTablePartTile:
    relative_start_time_inclusive: pendulum.Duration
    relative_end_time_exclusive: pendulum.Duration

    @classmethod
    def from_proto(cls, proto: schema_pb2.OnlineBatchTablePartTile) -> "OnlineBatchTablePartTile":
        return cls(
            relative_start_time_inclusive=time_utils.proto_to_duration(proto.relative_start_time_inclusive),
            relative_end_time_exclusive=time_utils.proto_to_duration(proto.relative_end_time_exclusive),
        )


@utils.frozen_strict
class OnlineBatchTablePart:
    window_index: int
    time_window: TimeWindowSpec
    schema: schema.Schema
    # We allow divergence between local and remote for tiles because we do not compute the tiles for args.
    tiles: Tuple[OnlineBatchTablePartTile, ...] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})

    @classmethod
    def from_proto(cls, proto: schema_pb2.OnlineBatchTablePart) -> "OnlineBatchTablePart":
        part_schema = _remove_deprecated_column_proto_fields(proto.schema)
        return cls(
            window_index=proto.window_index,
            time_window=create_time_window_spec_from_data_proto(proto.time_window),
            schema=schema.Schema(part_schema),
            tiles=tuple(OnlineBatchTablePartTile.from_proto(tile) for tile in proto.tiles),
        )


@utils.frozen_strict
class Inference:
    input_columns: List[schema.Column]
    output_column: schema.Column
    # TODO(jiadong): Make this a data class to wrap the proto
    model_artifact: model_artifact_service_pb2.ModelArtifactInfo

    @classmethod
    def from_data_proto(cls, inference_proto: feature_view__data_pb2.InferenceFeature) -> "Inference":
        return cls(
            input_columns=[schema.Column.from_proto(proto) for proto in inference_proto.input_columns],
            output_column=schema.Column.from_proto(inference_proto.output_column),
            model_artifact=inference_proto.model_artifact,
        )

    # TODO(jiadong): Connect this with MDS to get registered model artifact.
    @classmethod
    def from_args_proto(cls, _: feature_view__args_pb2.Inference) -> "Inference":
        msg = "`from_args_proto` for `Inference` has not been implemented yet"
        raise NotImplementedError(msg)


@utils.frozen_strict
class Embedding:
    input_column_name: str
    output_column_name: str
    model: str

    @classmethod
    def from_args_proto(
        cls,
        embedding_proto: feature_view__args_pb2.Embedding,
    ) -> "Embedding":
        return cls(
            input_column_name=embedding_proto.column,
            output_column_name=embedding_proto.name,
            model=embedding_proto.model,
        )

    @classmethod
    def from_data_proto(
        cls,
        embedding_proto: feature_view__data_pb2.EmbeddingFeature,
    ) -> "Embedding":
        return cls(
            input_column_name=embedding_proto.input_column_name,
            output_column_name=embedding_proto.output_column_name,
            model=embedding_proto.model,
        )


@utils.frozen_strict
class OnlineBatchTableFormat:
    online_batch_table_parts: Tuple[OnlineBatchTablePart, ...]

    @classmethod
    def from_proto(cls, proto: schema_pb2.OnlineBatchTableFormat) -> "OnlineBatchTableFormat":
        return cls(
            online_batch_table_parts=tuple(
                OnlineBatchTablePart.from_proto(part) for part in proto.online_batch_table_parts
            )
        )


class MaterializedFeatureViewType(enum.Enum):
    TEMPORAL = 1
    TEMPORAL_AGGREGATE = 2


@utils.frozen_strict
class MaterializedFeatureViewSpec(FeatureViewSpec):
    """Spec for Batch and Stream feature views."""

    is_continuous: bool
    type: MaterializedFeatureViewType
    data_source_type: data_source_type_pb2.DataSourceType.ValueType
    incremental_backfills: bool
    # TODO(TEC-19861): This should become non-optional once schemas mandatroy in plan/apply/test
    timestamp_field: Optional[str]

    # TODO(TEC-12321): Audit and fix feature view spec fields that should be required.
    pipeline: Optional[pipeline_pb2.Pipeline]

    batch_schedule: Optional[pendulum.Duration]
    slide_interval: Optional[pendulum.Duration]
    ttl: Optional[pendulum.Duration]
    feature_start_time: Optional[pendulum.DateTime]
    materialization_start_time: Optional[pendulum.DateTime]
    max_source_data_delay: pendulum.Duration
    materialized_data_path: Optional[str]
    published_features_path: Optional[str]
    time_range_policy: Optional[feature_view__data_pb2.MaterializationTimeRangePolicy.ValueType]
    materialization_state_transitions: Tuple[feature_view__data_pb2.MaterializationStateTransition, ...] = attrs.field(
        metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True}
    )

    offline_store: Optional[feature_view__args_pb2.OfflineFeatureStoreConfig]
    offline_store_params: Optional[feature_view__data_pb2.OfflineStoreParams] = attrs.field(
        metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True}
    )

    # Generally, data protos should not be exposed in the "spec". However, we make an exception in this case because
    # (a) there is no equivalent args proto, (b) it's a good data model for this usage, and (c) this proto is used
    # extensively in the query gen code (not worth refactoring).
    aggregate_features: Tuple[feature_view__data_pb2.AggregateFeature, ...]

    embedding_features: Tuple[Embedding, ...]
    inference_features: Tuple[Inference, ...]

    aggregation_secondary_key: Optional[str]
    secondary_key_rollup_outputs: Optional[Tuple[SecondaryKeyOutputColumn, ...]]

    slide_interval_string: Optional[str]

    # Only relevant for offline-materialized fvs on snowflake compute
    snowflake_view_name: Optional[str] = attrs.field(metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True})

    # TODO(TEC-12321): Audit and fix feature view spec fields that should be required. (batch_cluster_config should be.)
    batch_cluster_config: Optional[feature_view__args_pb2.ClusterConfig]
    stream_cluster_config: Optional[feature_view__args_pb2.ClusterConfig]

    batch_compute_mode: BatchComputeMode

    # TODO(TEC-12321): Audit and fix feature view spec fields that should be required.
    # See failure: https://tectonworkspace.slack.com/archives/C04L8M14XGX/p1675279851469019
    batch_trigger: Optional[feature_view__args_pb2.BatchTriggerType.ValueType]
    manual_trigger_backfill_end_time: Optional[pendulum.DateTime]

    online_batch_table_format: Optional[OnlineBatchTableFormat]
    compaction_enabled: bool
    stream_tiling_enabled: bool

    has_explicit_view_schema: bool

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: feature_view__data_pb2.FeatureView) -> "MaterializedFeatureViewSpec":
        if proto.HasField("temporal_aggregate"):
            fv_type = MaterializedFeatureViewType.TEMPORAL_AGGREGATE
            is_continuous = proto.temporal_aggregate.is_continuous
            data_source_type = utils.get_field_or_none(proto.temporal_aggregate, "data_source_type")
            incremental_backfills = False
            slide_interval = time_utils.proto_to_duration(proto.temporal_aggregate.slide_interval)
            ttl = None
            # TODO(sanika): add a wrapper for aggregate features spec
            aggregate_features = utils.get_tuple_from_repeated_field(proto.temporal_aggregate.features)
            embedding_features = ()
            inference_features = ()
            slide_interval_string = utils.get_field_or_none(proto.temporal_aggregate, "slide_interval_string")
            aggregation_secondary_key = utils.get_field_or_none(proto.temporal_aggregate, "aggregation_secondary_key")
            secondary_key_rollup_outputs = None
            if aggregation_secondary_key:
                secondary_key_rollup_outputs = tuple(
                    SecondaryKeyOutputColumn(
                        name=output_col.name,
                        time_window=create_time_window_spec_from_data_proto(output_col.time_window),
                    )
                    for output_col in proto.temporal_aggregate.secondary_key_output_columns
                )
        elif proto.HasField("temporal"):
            fv_type = MaterializedFeatureViewType.TEMPORAL
            is_continuous = proto.temporal.is_continuous
            data_source_type = utils.get_field_or_none(proto.temporal, "data_source_type")
            incremental_backfills = proto.temporal.incremental_backfills
            slide_interval = None
            ttl = utils.get_non_default_duration_field_or_none(proto.temporal, "serving_ttl")
            aggregate_features = ()
            embedding_features = tuple(Embedding.from_data_proto(emb_proto) for emb_proto in proto.temporal.embeddings)
            inference_features = tuple(
                Inference.from_data_proto(inference_proto) for inference_proto in proto.temporal.inferences
            )
            slide_interval_string = None
            aggregation_secondary_key = None
            secondary_key_rollup_outputs = None
        else:
            msg = f"Unexpected feature view type: {proto}"
            raise TypeError(msg)

        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.feature_view_id, proto.fco_metadata
            ),
            entity_ids=tuple(id_helper.IdHelper.to_string(id) for id in proto.entity_ids),
            join_keys=utils.get_tuple_from_repeated_field(proto.join_keys),
            online_serving_keys=utils.get_tuple_from_repeated_field(proto.online_serving_index.join_keys),
            view_schema=_get_view_schema(proto.schemas),
            materialization_schema=_get_materialization_schema(proto.schemas),
            has_explicit_view_schema=proto.schemas.is_explicit_view_schema,
            offline_store=utils.get_field_or_none(proto.materialization_params, "offline_store_config"),
            offline_store_params=utils.get_field_or_none(proto.materialization_params, "offline_store_params"),
            is_continuous=is_continuous,
            data_source_type=data_source_type,
            incremental_backfills=incremental_backfills,
            timestamp_field=utils.get_field_or_none(proto, "timestamp_key"),
            type=fv_type,
            feature_store_format_version=proto.feature_store_format_version,
            materialization_enabled=proto.materialization_enabled,
            online=proto.materialization_params.writes_to_online_store,
            offline=proto.materialization_params.writes_to_offline_store,
            pipeline=utils.get_field_or_none(proto, "pipeline"),
            batch_schedule=utils.get_non_default_duration_field_or_none(
                proto.materialization_params, "schedule_interval"
            ),
            slide_interval=slide_interval,
            ttl=ttl,
            feature_start_time=utils.get_timestamp_field_or_none(
                proto.materialization_params, "feature_start_timestamp"
            ),
            materialization_start_time=utils.get_timestamp_field_or_none(
                proto.materialization_params, "materialization_start_timestamp"
            ),
            max_source_data_delay=time_utils.proto_to_duration(proto.materialization_params.max_source_data_delay),
            aggregate_features=aggregate_features,
            aggregation_secondary_key=aggregation_secondary_key,
            secondary_key_rollup_outputs=secondary_key_rollup_outputs,
            embedding_features=embedding_features,
            inference_features=inference_features,
            slide_interval_string=slide_interval_string,
            materialized_data_path=utils.get_field_or_none(
                proto.enrichments.fp_materialization.materialized_data_location, "path"
            ),
            published_features_path=utils.get_field_or_none(
                proto.enrichments.fp_materialization.feature_export_data_location, "path"
            ),
            materialization_state_transitions=utils.get_tuple_from_repeated_field(
                proto.materialization_state_transitions
            ),
            time_range_policy=utils.get_field_or_none(proto.materialization_params, "time_range_policy"),
            snowflake_view_name=utils.get_field_or_none(proto.snowflake_data, "snowflake_view_name"),
            validation_args=validator_pb2.FcoValidationArgs(feature_view=proto.validation_args),
            batch_cluster_config=utils.get_field_or_none(proto.materialization_params, "batch_materialization"),
            stream_cluster_config=utils.get_field_or_none(proto.materialization_params, "stream_materialization"),
            batch_trigger=utils.get_field_or_none(proto, "batch_trigger"),
            manual_trigger_backfill_end_time=utils.get_timestamp_field_or_none(
                proto.materialization_params, "manual_trigger_backfill_end_timestamp"
            ),
            url=utils.get_field_or_none(proto, "web_url"),
            online_batch_table_format=_get_online_batch_table_format(proto.schemas),
            batch_compute_mode=BatchComputeMode(proto.batch_compute_mode)
            if proto.HasField("batch_compute_mode")
            else None,
            compaction_enabled=proto.materialization_params.compaction_enabled,
            stream_tiling_enabled=proto.materialization_params.stream_tiling_enabled,
            options=MappingProxyType(proto.options),
            context_parameter_name=proto.context_parameter_name,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: feature_view__args_pb2.FeatureViewArgs, supplement: FeatureViewSpecArgsSupplement
    ) -> "MaterializedFeatureViewSpec":
        feature_start_time = utils.get_timestamp_field_or_none(
            proto.materialized_feature_view_args, "feature_start_time"
        )

        compaction_enabled = proto.materialized_feature_view_args.compaction_enabled
        is_continuous = get_is_continuous_from_feature_view_args(proto)
        data_source_type = utils.get_field_or_none(proto.materialized_feature_view_args, "data_source_type")
        aggregate_features = get_aggregate_features_from_feature_view_args(proto)
        embedding_features = tuple(
            Embedding.from_args_proto(emb_args_proto)
            for emb_args_proto in proto.materialized_feature_view_args.embeddings
        )

        is_aggregate = len(proto.materialized_feature_view_args.aggregations) > 0
        if is_aggregate:
            fv_type = MaterializedFeatureViewType.TEMPORAL_AGGREGATE

            slide_interval_proto = _get_aggregation_interval_from_feature_view_args(
                proto, is_continuous, data_source_type
            )
            if slide_interval_proto:
                slide_interval = time_utils.proto_to_duration(slide_interval_proto)
                slide_interval_string = feature_view_utils.construct_aggregation_interval_name(
                    slide_interval_proto, is_continuous
                )
            else:
                slide_interval = None
                slide_interval_string = ""

            lifetime_start_time = utils.get_timestamp_field_or_none(
                proto.materialized_feature_view_args, "lifetime_start_time"
            )
            if lifetime_start_time:
                materialization_start_time = lifetime_start_time
            elif feature_start_time is not None:
                # Logic must be kept in sync with getMaterializationStartTime() in FeatureViewManager.
                earliest_window_start = min(
                    [
                        RelativeTimeWindowSpec.from_args_proto(agg.time_window).window_start
                        for agg in proto.materialized_feature_view_args.aggregations
                    ]
                )
                materialization_start_time = feature_start_time + earliest_window_start
            else:
                materialization_start_time = None

            aggregation_secondary_key = utils.get_field_or_none(
                proto.materialized_feature_view_args, "aggregation_secondary_key"
            )
            secondary_key_rollup_outputs = None
            if aggregation_secondary_key:
                secondary_key_rollup_outputs = tuple(
                    SecondaryKeyOutputColumn(
                        name=output_col.name, time_window=args_oneof_timewindow_to_spec(output_col)
                    )
                    for output_col in proto.materialized_feature_view_args.secondary_key_output_columns
                )
        else:
            fv_type = MaterializedFeatureViewType.TEMPORAL
            slide_interval_string = None
            materialization_start_time = feature_start_time
            slide_interval = None
            aggregation_secondary_key = None
            secondary_key_rollup_outputs = None

        try:
            timestamp_field = resolve_timestamp_field(proto, supplement.view_schema)
        except errors.TectonValidationError:
            # TODO(TEC-19861): Remove when we schemas mandatroy in plan/apply/test
            if not conf.get_bool("TECTON_REQUIRE_SCHEMA"):
                timestamp_field = None
            else:
                raise

        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(proto.feature_view_id, proto.info),
            aggregation_secondary_key=aggregation_secondary_key,
            secondary_key_rollup_outputs=secondary_key_rollup_outputs,
            entity_ids=tuple(id_helper.IdHelper.to_string(entity.entity_id) for entity in proto.entities),
            join_keys=get_join_keys_from_feature_view_args(proto),
            online_serving_keys=get_online_serving_keys_from_feature_view_args(proto),
            view_schema=schema.Schema(supplement.view_schema),
            materialization_schema=schema.Schema(supplement.materialization_schema),
            has_explicit_view_schema=proto.materialized_feature_view_args.HasField("schema"),
            offline_store=utils.get_field_or_none(
                proto.materialized_feature_view_args.offline_store, "staging_table_format"
            ),
            offline_store_params=None,
            is_continuous=is_continuous,
            data_source_type=data_source_type,
            incremental_backfills=proto.materialized_feature_view_args.incremental_backfills,
            timestamp_field=timestamp_field,
            type=fv_type,
            feature_store_format_version=feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TIME_NANOSECONDS,
            materialization_enabled=False,
            online=proto.online_enabled,
            offline=proto.offline_enabled,
            pipeline=utils.get_field_or_none(proto, "pipeline"),
            batch_schedule=get_batch_schedule_from_feature_view_args(proto),
            slide_interval=slide_interval,
            ttl=utils.get_duration_field_or_none(proto.materialized_feature_view_args, "serving_ttl"),
            feature_start_time=feature_start_time,
            materialization_start_time=materialization_start_time,
            max_source_data_delay=_get_max_schedule_offset(proto.pipeline),
            aggregate_features=aggregate_features,
            embedding_features=embedding_features,
            # TODO(jiadong): Support local dev for Inference feature.
            inference_features=(),
            slide_interval_string=slide_interval_string,
            materialized_data_path=None,
            published_features_path=None,
            time_range_policy=feature_view__data_pb2.MaterializationTimeRangePolicy.MATERIALIZATION_TIME_RANGE_POLICY_FILTER_TO_RANGE,
            materialization_state_transitions=(),
            snowflake_view_name=None,
            validation_args=None,
            batch_cluster_config=utils.get_field_or_none(proto.materialized_feature_view_args, "batch_compute"),
            stream_cluster_config=utils.get_field_or_none(proto.materialized_feature_view_args, "stream_compute"),
            batch_trigger=get_batch_trigger_from_feature_view_args(proto),
            manual_trigger_backfill_end_time=utils.get_timestamp_field_or_none(
                proto.materialized_feature_view_args, "manual_trigger_backfill_end_time"
            ),
            url=None,
            online_batch_table_format=OnlineBatchTableFormat.from_proto(supplement.online_batch_table_format)
            if supplement.online_batch_table_format
            else None,
            batch_compute_mode=BatchComputeMode(proto.batch_compute_mode)
            if proto.HasField("batch_compute_mode")
            else None,
            compaction_enabled=compaction_enabled,
            stream_tiling_enabled=proto.materialized_feature_view_args.stream_tiling_enabled,
            options=MappingProxyType(proto.options),
            context_parameter_name=proto.context_parameter_name,
        )

    @property
    def features(self) -> List[str]:
        if len(self.aggregate_features) > 0:
            # Temporal aggregate feature view.
            secondary_key_output_cols = (
                [col.name for col in self.secondary_key_rollup_outputs] if self.secondary_key_rollup_outputs else []
            )
            aggregate_output_cols = [
                aggregate_feature.output_feature_name for aggregate_feature in self.aggregate_features
            ]
            return secondary_key_output_cols + aggregate_output_cols
        else:
            # Temporal feature view.
            non_feature_cols = {self.timestamp_field, *self.join_keys, "_anchor_time", "_ANCHOR_TIME"}
            cols = [col for col in self.materialization_schema.column_names() if col not in non_feature_cols]
            return cols


@utils.frozen_strict
class RealtimeFeatureViewSpec(FeatureViewSpec):
    # TODO(TEC-12321): Audit and fix feature view spec fields that should be required.
    pipeline: Optional[pipeline_pb2.Pipeline]

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: feature_view__data_pb2.FeatureView) -> "RealtimeFeatureViewSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.feature_view_id, proto.fco_metadata
            ),
            entity_ids=tuple(id_helper.IdHelper.to_string(id) for id in proto.entity_ids),
            join_keys=utils.get_tuple_from_repeated_field(proto.join_keys),
            online_serving_keys=utils.get_tuple_from_repeated_field(proto.online_serving_index.join_keys),
            view_schema=_get_view_schema(proto.schemas),
            materialization_schema=_get_materialization_schema(proto.schemas),
            feature_store_format_version=proto.feature_store_format_version,
            materialization_enabled=False,
            online=False,
            offline=False,
            pipeline=utils.get_field_or_none(proto, "pipeline"),
            validation_args=validator_pb2.FcoValidationArgs(feature_view=proto.validation_args),
            url=utils.get_field_or_none(proto, "web_url"),
            options=MappingProxyType(proto.options),
            context_parameter_name=proto.context_parameter_name,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: feature_view__args_pb2.FeatureViewArgs, supplement: FeatureViewSpecArgsSupplement
    ) -> "RealtimeFeatureViewSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(proto.feature_view_id, proto.info),
            entity_ids=(),
            join_keys=(),
            online_serving_keys=(),
            view_schema=schema.Schema(supplement.view_schema),
            materialization_schema=schema.Schema(supplement.materialization_schema),
            feature_store_format_version=feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TIME_NANOSECONDS,
            materialization_enabled=False,
            online=False,
            offline=False,
            pipeline=utils.get_field_or_none(proto, "pipeline"),
            validation_args=None,
            url=None,
            options=MappingProxyType(proto.options),
            context_parameter_name=proto.context_parameter_name,
        )

    @property
    def features(self) -> List[str]:
        return list(self.view_schema.column_names())

    def join_key_schema(
        self, dependent_fv_specs_and_jk_overrides: List[Tuple[FeatureViewSpec, List[utils.JoinKeyMappingSpec]]]
    ) -> schema.Schema:
        """Returns the combined join key schema from all input FVs which are adjusted by join key overridings."""
        input_fv_schema = schema.Schema(schema_pb2.Schema())
        for fv_spec, jk_overrides in dependent_fv_specs_and_jk_overrides:
            fv_schema_dict = fv_spec.join_key_schema().to_dict()
            for jk_override_spec in jk_overrides:
                fv_schema_dict[jk_override_spec.spine_column_name] = fv_schema_dict[
                    jk_override_spec.feature_view_column_name
                ]
                # Delete the original feature_view_column_name entry as the data type is assigned to the overriding key.
                del fv_schema_dict[jk_override_spec.feature_view_column_name]
            input_fv_schema += schema.Schema.from_dict(fv_schema_dict)
        return input_fv_schema


@utils.frozen_strict
class FeatureTableSpec(FeatureViewSpec):
    timestamp_field: str
    ttl: Optional[pendulum.Duration]

    offline_store: Optional[feature_view__args_pb2.OfflineFeatureStoreConfig]
    offline_store_params: Optional[feature_view__data_pb2.OfflineStoreParams] = attrs.field(
        metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True}
    )
    materialized_data_path: Optional[str]
    time_range_policy: Optional[feature_view__data_pb2.MaterializationTimeRangePolicy.ValueType]
    materialization_state_transitions: Tuple[feature_view__data_pb2.MaterializationStateTransition, ...] = attrs.field(
        metadata={utils.LOCAL_REMOTE_DIVERGENCE_ALLOWED: True}
    )

    # TODO(TEC-12321): Audit and fix feature view spec fields that should be required. (batch_cluster_config should be.)
    batch_cluster_config: Optional[feature_view__args_pb2.ClusterConfig]

    @classmethod
    @typechecked
    def from_data_proto(cls, proto: feature_view__data_pb2.FeatureView) -> "FeatureTableSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_data_proto(
                proto.feature_view_id, proto.fco_metadata
            ),
            entity_ids=tuple(id_helper.IdHelper.to_string(id) for id in proto.entity_ids),
            join_keys=utils.get_tuple_from_repeated_field(proto.join_keys),
            online_serving_keys=utils.get_tuple_from_repeated_field(proto.online_serving_index.join_keys),
            view_schema=_get_view_schema(proto.schemas),
            materialization_schema=_get_materialization_schema(proto.schemas),
            offline_store=utils.get_field_or_none(proto.materialization_params, "offline_store_config"),
            offline_store_params=utils.get_field_or_none(proto.materialization_params, "offline_store_params"),
            timestamp_field=utils.get_field_or_none(proto, "timestamp_key"),
            feature_store_format_version=proto.feature_store_format_version,
            materialization_enabled=proto.materialization_enabled,
            online=proto.feature_table.online_enabled,
            offline=proto.feature_table.offline_enabled,
            ttl=utils.get_non_default_duration_field_or_none(proto.feature_table, "serving_ttl"),
            materialized_data_path=utils.get_field_or_none(
                proto.enrichments.fp_materialization.materialized_data_location, "path"
            ),
            materialization_state_transitions=utils.get_tuple_from_repeated_field(
                proto.materialization_state_transitions
            ),
            time_range_policy=utils.get_field_or_none(proto.materialization_params, "time_range_policy"),
            validation_args=validator_pb2.FcoValidationArgs(feature_view=proto.validation_args),
            batch_cluster_config=utils.get_field_or_none(proto.materialization_params, "batch_materialization"),
            url=utils.get_field_or_none(proto, "web_url"),
            options=MappingProxyType(proto.options),
            context_parameter_name=None,
        )

    @classmethod
    @typechecked
    def from_args_proto(
        cls, proto: feature_view__args_pb2.FeatureViewArgs, supplement: FeatureViewSpecArgsSupplement
    ) -> "FeatureTableSpec":
        return cls(
            metadata=tecton_object_spec.TectonObjectMetadataSpec.from_args_proto(proto.feature_view_id, proto.info),
            entity_ids=tuple(id_helper.IdHelper.to_string(entity.entity_id) for entity in proto.entities),
            join_keys=get_join_keys_from_feature_view_args(proto),
            online_serving_keys=get_online_serving_keys_from_feature_view_args(proto),
            view_schema=schema.Schema(supplement.view_schema),
            materialization_schema=schema.Schema(supplement.materialization_schema),
            offline_store=utils.get_field_or_none(proto.feature_table_args.offline_store, "staging_table_format"),
            offline_store_params=None,
            timestamp_field=_get_timestamp_column(supplement.view_schema),
            feature_store_format_version=feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_TIME_NANOSECONDS,
            materialization_enabled=False,
            online=proto.online_enabled,
            offline=proto.offline_enabled,
            ttl=utils.get_duration_field_or_none(proto.feature_table_args, "serving_ttl"),
            materialized_data_path=None,
            materialization_state_transitions=(),
            time_range_policy=None,
            validation_args=None,
            batch_cluster_config=utils.get_field_or_none(proto.feature_table_args, "batch_compute"),
            url=None,
            options=MappingProxyType(proto.options),
            context_parameter_name=None,
        )

    @property
    def features(self) -> List[str]:
        return [
            col for col in self.view_schema.column_names() if col != self.timestamp_field and col not in self.join_keys
        ]


def _remove_deprecated_column_proto_fields(schema: schema_pb2.Schema) -> schema_pb2.Schema:
    """Clear out deprecated fields from the column proto

    The deprecated fields are still served in data proto for backwards compatibility with old SDKs.
    Clear the field here so that:
        1) Developers don't depend on this field being set.
        2) Local specs are equivalent to the remote specs.
    """

    # Copy the schema so the `ClearField` doesn't change the proto passed into this fuction such as the schema in the MaterializationTaskParam.
    schema_copy = copy.deepcopy(schema)
    for column in schema_copy.columns:
        column.ClearField("feature_server_type")
        column.ClearField("raw_spark_type")
        column.ClearField("raw_snowflake_type")
    return schema_copy


def _get_view_schema(schemas: feature_view__data_pb2.FeatureViewSchemas) -> Optional[schema.Schema]:
    if not schemas.HasField("view_schema"):
        return None

    view_schema = _remove_deprecated_column_proto_fields(schemas.view_schema)
    return schema.Schema(view_schema)


def _get_materialization_schema(schemas: feature_view__data_pb2.FeatureViewSchemas) -> Optional[schema.Schema]:
    if not schemas.HasField("materialization_schema"):
        return None

    materialization_schema = _remove_deprecated_column_proto_fields(schemas.materialization_schema)
    return schema.Schema(materialization_schema)


def _get_online_batch_table_format(
    schemas: feature_view__data_pb2.FeatureViewSchemas,
) -> Optional[OnlineBatchTableFormat]:
    if not schemas.HasField("online_batch_table_format"):
        return None

    return OnlineBatchTableFormat.from_proto(schemas.online_batch_table_format)


@typechecked
def create_feature_view_spec_from_data_proto(
    proto: feature_view__data_pb2.FeatureView,
) -> Optional[Union[MaterializedFeatureViewSpec, RealtimeFeatureViewSpec, FeatureTableSpec]]:
    if proto.HasField("temporal_aggregate") or proto.HasField("temporal"):
        return MaterializedFeatureViewSpec.from_data_proto(proto)
    elif proto.HasField("realtime_feature_view"):
        return RealtimeFeatureViewSpec.from_data_proto(proto)
    elif proto.HasField("feature_table"):
        return FeatureTableSpec.from_data_proto(proto)
    else:
        msg = f"Unexpect feature view type: {proto}"
        raise ValueError(msg)


@typechecked
def create_feature_view_spec_from_args_proto(
    proto: feature_view__args_pb2.FeatureViewArgs,
    supplement: FeatureViewSpecArgsSupplement,
) -> Optional[Union[MaterializedFeatureViewSpec, RealtimeFeatureViewSpec, FeatureTableSpec]]:
    if proto.HasField("materialized_feature_view_args"):
        return MaterializedFeatureViewSpec.from_args_proto(proto, supplement)
    elif proto.HasField("realtime_args"):
        return RealtimeFeatureViewSpec.from_args_proto(proto, supplement)
    elif proto.HasField("feature_table_args"):
        return FeatureTableSpec.from_args_proto(proto, supplement)
    else:
        msg = f"Unexpect feature view type: {proto}"
        raise ValueError(msg)


def _get_timestamp_column(schema_proto: schema_pb2.Schema) -> str:
    schema_ = schema.Schema(schema_proto)
    timestamp_columns = [
        column[0] for column in schema_.column_name_and_data_types() if column[1] == tecton_types.TimestampType()
    ]
    if len(timestamp_columns) != 1:
        msg = f"Attempted to infer timestamp. Expected exactly one timestamp column in schema {schema_}"
        raise errors.TectonValidationError(msg)
    return timestamp_columns[0]


def _get_max_schedule_offset(pipeline: pipeline_pb2.Pipeline) -> pendulum.Duration:
    ds_nodes = pipeline_common.get_all_data_source_nodes(pipeline)
    assert len(ds_nodes) > 0
    return max([time_utils.proto_to_duration(ds_node.data_source_node.schedule_offset) for ds_node in ds_nodes])


@typechecked
def get_batch_schedule_from_feature_view_args(
    proto: feature_view__args_pb2.FeatureViewArgs,
) -> Optional[pendulum.Duration]:
    is_aggregate = len(proto.materialized_feature_view_args.aggregations) > 0
    if is_aggregate:
        slide_interval = utils.get_duration_field_or_none(proto.materialized_feature_view_args, "aggregation_interval")
        is_continuous = (
            proto.materialized_feature_view_args.stream_processing_mode
            == feature_view__args_pb2.StreamProcessingMode.STREAM_PROCESSING_MODE_CONTINUOUS
        )

        if is_continuous:
            # Default is set in Kotlin to one day per CONTINUOUS_MODE_TILE_DURATION.
            return pendulum.Duration(days=1)
        else:
            return (
                utils.get_duration_field_or_none(proto.materialized_feature_view_args, "batch_schedule")
                or slide_interval
            )
    else:
        return utils.get_duration_field_or_none(proto.materialized_feature_view_args, "batch_schedule")


# TODO: fix/cleanup
def args_oneof_timewindow_to_spec(proto: Any) -> TimeWindowSpec:  # noqa: ANN401
    if proto.HasField("time_window"):
        return RelativeTimeWindowSpec.from_args_proto(proto.time_window)
    elif proto.HasField("time_window_series"):
        return TimeWindowSeriesSpec.from_args_proto(proto.time_window_series)
    elif proto.HasField("lifetime_window"):
        return LifetimeWindowSpec()
    else:
        msg = f"Unexpected time window type in agg args proto: {proto}"
        raise ValueError(msg)


@typechecked
def get_aggregate_features_from_feature_view_args(
    proto: feature_view__args_pb2.FeatureViewArgs,
) -> Tuple[feature_view__data_pb2.AggregateFeature, ...]:
    is_continuous = (
        proto.materialized_feature_view_args.stream_processing_mode
        == feature_view__args_pb2.StreamProcessingMode.STREAM_PROCESSING_MODE_CONTINUOUS
    )
    aggregate_features = []
    for agg_args_proto in proto.materialized_feature_view_args.aggregations:
        window_spec = args_oneof_timewindow_to_spec(agg_args_proto)

        agg_data_proto = create_aggregate_features(
            agg_args_proto, proto.materialized_feature_view_args.aggregation_interval, is_continuous, window_spec
        )
        aggregate_features.append(agg_data_proto)
    return tuple(aggregate_features)


@typechecked
def _get_aggregation_interval_from_feature_view_args(
    proto: feature_view__args_pb2.FeatureViewArgs,
    is_continuous: bool,
    data_source_type: Optional[data_source_type_pb2.DataSourceType.ValueType],
) -> Optional[duration_pb2.Duration]:
    if proto.materialized_feature_view_args.compaction_enabled:
        is_stream = (
            data_source_type == data_source_type_pb2.DataSourceType.STREAM_WITH_BATCH if data_source_type else False
        )
        # Default set in Kotlin for batch fvs with compaction is the batch schedule. Batch schedule is required for fvs with compaction_enabled=True
        # Default set in Kotlin for non-continuous stream fvs with compaction is the stream compaction tile size.
        if not is_stream:
            return utils.get_field_or_none(proto.materialized_feature_view_args, "batch_schedule")
        elif not is_continuous:
            return utils.get_field_or_none(proto.materialized_feature_view_args, "stream_tile_size")

    return utils.get_field_or_none(proto.materialized_feature_view_args, "aggregation_interval")


@typechecked
def get_is_continuous_from_feature_view_args(
    proto: feature_view__args_pb2.FeatureViewArgs,
) -> bool:
    data_source_type = utils.get_field_or_none(proto.materialized_feature_view_args, "data_source_type")
    if not data_source_type or data_source_type == data_source_type_pb2.DataSourceType.BATCH:
        return False

    if proto.materialized_feature_view_args.compaction_enabled:
        return proto.materialized_feature_view_args.stream_tiling_enabled == False

    return (
        proto.materialized_feature_view_args.stream_processing_mode
        == feature_view__args_pb2.StreamProcessingMode.STREAM_PROCESSING_MODE_CONTINUOUS
    )


@typechecked
def get_batch_trigger_from_feature_view_args(
    proto: feature_view__args_pb2.FeatureViewArgs,
) -> Optional[feature_view__args_pb2.BatchTriggerType.ValueType]:
    if proto.materialized_feature_view_args.HasField("batch_trigger"):
        return proto.materialized_feature_view_args.batch_trigger
    else:
        return feature_view__args_pb2.BatchTriggerType.BATCH_TRIGGER_TYPE_SCHEDULED


@typechecked
def get_join_keys_from_feature_view_args(proto: feature_view__args_pb2.FeatureViewArgs) -> Tuple[str, ...]:
    join_keys = []
    for entity in proto.entities:
        join_keys.extend(entity.join_keys)
    return tuple(join_keys)


@typechecked
def resolve_timestamp_field(
    feature_view_args: feature_view__args_pb2.FeatureViewArgs,
    view_schema: schema_pb2.Schema,
) -> str:
    if feature_view_args.materialized_feature_view_args.HasField("timestamp_field"):
        return feature_view_args.materialized_feature_view_args.timestamp_field
    return _get_timestamp_column(view_schema)


@typechecked
def get_online_serving_keys_from_feature_view_args(proto: feature_view__args_pb2.FeatureViewArgs) -> Tuple[str, ...]:
    return (
        tuple(proto.online_serving_index) if proto.online_serving_index else get_join_keys_from_feature_view_args(proto)
    )


@typechecked
def create_aggregate_features(
    feature_aggregation: feature_view__args_pb2.FeatureAggregation,
    aggregation_interval_seconds: duration_pb2.Duration,
    is_continuous: bool,
    time_window: TimeWindowSpec,
) -> feature_view__data_pb2.AggregateFeature:
    """Build a AggregateFeature data proto from the input FeatureAggregation args proto."""
    feature_function = get_aggregation_enum_from_string(feature_aggregation.function.lower())
    assert feature_function, f"Unknown aggregation name: {feature_aggregation.function}"

    if feature_function in {
        afpb.AggregationFunction.AGGREGATION_FUNCTION_LAST_DISTINCT_N,
        afpb.AggregationFunction.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N,
    }:
        function_params = afpb.AggregationFunctionParams(
            last_n=afpb.LastNParams(n=feature_aggregation.function_params["n"].int64_value)
        )
    elif feature_function in {
        afpb.AggregationFunction.AGGREGATION_FUNCTION_FIRST_DISTINCT_N,
        afpb.AggregationFunction.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N,
    }:
        function_params = afpb.AggregationFunctionParams(
            first_n=afpb.FirstNParams(n=feature_aggregation.function_params["n"].int64_value)
        )
    elif feature_function == afpb.AggregationFunction.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT:
        function_params = afpb.AggregationFunctionParams(
            approx_count_distinct=afpb.ApproxCountDistinctParams(
                precision=feature_aggregation.function_params["precision"].int64_value
            )
        )
    elif feature_function == afpb.AggregationFunction.AGGREGATION_FUNCTION_APPROX_PERCENTILE:
        function_params = afpb.AggregationFunctionParams(
            approx_percentile=afpb.ApproxPercentileParams(
                precision=feature_aggregation.function_params["precision"].int64_value,
                percentile=feature_aggregation.function_params["percentile"].double_value,
            )
        )
    else:
        function_params = None

    if len(feature_aggregation.function_params) > 0:
        assert (
            function_params is not None
        ), "function_params in the data proto should not be None since it is a non-empty dictionary in the args proto"
    return feature_view__data_pb2.AggregateFeature(
        input_feature_name=feature_aggregation.column,
        output_feature_name=feature_aggregation.name,
        function=feature_function,
        time_window=window_spec_to_window_data_proto(time_window),
        function_params=function_params,
        batch_sawtooth_tile_size=feature_aggregation.batch_sawtooth_tile_size
        if feature_aggregation.HasField("batch_sawtooth_tile_size")
        else None,
    )



import dataclasses
import functools
import json
import os
import typing
import uuid
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from urllib.parse import urlparse

import pyarrow
import pyarrow.dataset
import pyarrow.fs
import ray
from deltalake.writer import DeltaJSONEncoder
from deltalake.writer import get_file_stats_from_metadata
from pyarrow._dataset import WrittenFile

from tecton_core import offline_store
from tecton_core.arrow import PARQUET_WRITE_OPTIONS
from tecton_core.compute_mode import ComputeMode
from tecton_core.duckdb_context import DuckDBContext
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.offline_store import patch_timestamps_in_arrow_schema
from tecton_core.query.dialect import Dialect
from tecton_core.query.nodes import AddAnchorTimeNode
from tecton_core.query.nodes import ConvertTimestampToUTCNode
from tecton_core.query.nodes import StagedTableScanNode
from tecton_materialization.common.task_params import TimeInterval
from tecton_materialization.ray.nodes import AddTimePartitionNode
from tecton_materialization.ray.nodes import TimeSpec
from tecton_proto.common import data_type__client_pb2 as data_type_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_proto.common.aws_credentials__client_pb2 import AwsIamRole
from tecton_proto.offlinestore.delta import metadata__client_pb2 as metadata_pb2
from tecton_proto.offlinestore.delta import transaction_writer__client_pb2 as transaction_writer_pb2
from tecton_proto.offlinestore.delta.metadata__client_pb2 import TectonDeltaMetadata


R = typing.TypeVar("R")
TxnFn = Callable[[], R]

Z_ORDER_TAG = "optimizationType"
Z_ORDER_TAG_VALUE = "z-order"
# A single Parquet Row Group (1_000_000 rows by default) is usually 16-32mb
# Our target size for a single Parquet file is 512-1024mb.
# Hence, the limit for number of rows per file:
PARQUET_MAX_ROWS_PER_FILE = 32_000_000


@dataclasses.dataclass
class OfflineStoreParams:
    feature_view_id: str
    feature_view_name: str
    schema: schema_pb2.Schema
    time_spec: TimeSpec
    feature_store_format_version: int
    batch_schedule: Optional[int]

    @staticmethod
    def for_feature_definition(fd: FeatureDefinitionWrapper) -> "OfflineStoreParams":
        return OfflineStoreParams(
            feature_view_id=fd.id,
            feature_view_name=fd.name,
            schema=fd.materialization_schema.to_proto(),
            time_spec=TimeSpec.for_feature_definition(fd),
            feature_store_format_version=fd.get_feature_store_format_version,
            # feature tables do not have schedules
            batch_schedule=fd.get_batch_schedule_for_version if not fd.is_feature_table else None,
        )


class DeltaConcurrentModificationException(Exception):
    def __init__(self, error_type: transaction_writer_pb2.UpdateResult.ErrorType, error_message: str):
        self.error_type = error_type
        self.error_message = error_message

    def __str__(self):
        error_type_name = transaction_writer_pb2.UpdateResult.ErrorType.Name(self.error_type)
        return (
            f"Delta commit failed due to a transaction conflict. "
            f"Conflict type: {error_type_name}. Message: {self.error_message}"
        )


class JavaActorWrapper:
    """Blocking wrapper around a Java actor."""

    def __init__(self, class_name):
        self.actor = ray.cross_language.java_actor_class(class_name).remote()

    def __getattr__(self, item):
        def f(*args):
            return ray.get(getattr(self.actor, item).remote(*args))

        return f


class TransactionWriter:
    """Wrapper around TransactionWriter actor which handles (de)serialization of parameters and return values."""

    def __init__(self, args: transaction_writer_pb2.InitializeArgs):
        self.actor = JavaActorWrapper("com.tecton.offlinestore.delta.TransactionWriterActor")
        self.actor.initialize(args.SerializeToString())

    def has_commit_with_metadata(self, metadata: metadata_pb2.TectonDeltaMetadata) -> bool:
        return self.actor.hasCommitWithMetadata(metadata.SerializeToString())

    def read_for_update(self, predicate: transaction_writer_pb2.Expression) -> List[str]:
        result_bytes = self.actor.readForUpdate(
            transaction_writer_pb2.ReadForUpdateArgs(read_predicate=predicate).SerializeToString()
        )
        result = transaction_writer_pb2.ReadForUpdateResult()
        result.ParseFromString(result_bytes)
        return result.uris

    def update(self, args: transaction_writer_pb2.UpdateArgs) -> transaction_writer_pb2.UpdateResult:
        result_bytes = self.actor.update(args.SerializeToString())
        result = transaction_writer_pb2.UpdateResult()
        result.ParseFromString(result_bytes)
        if not result.committed_version:
            raise DeltaConcurrentModificationException(error_type=result.error_type, error_message=result.error_message)

        return result

    def get_partitions_not_marked_with(self, tags: Dict[str, str]) -> List[Dict[str, str]]:
        args = transaction_writer_pb2.GetPartitionsArgs(tags=tags)
        result_bytes = self.actor.getPartitionsNotMarkedWith(args.SerializeToString())
        result = transaction_writer_pb2.GetPartitionsResult()
        result.ParseFromString(result_bytes)
        return [partition.values for partition in result.partitions]


def _pyarrow_literal(table: pyarrow.Table, column: schema_pb2.Column, row: int) -> transaction_writer_pb2.Expression:
    """Returns a Delta literal Expression for the given row and column within table."""
    pa_column = next((c for (name, c) in zip(table.column_names, table.columns) if name == column.name))
    pa_value = pa_column[row]

    def assert_type(t: pyarrow.DataType):
        assert pa_column.type == t, f"Type error for {column.name}. Expected {t}; got {pa_column.type}"

    if column.offline_data_type.type == data_type_pb2.DataTypeEnum.DATA_TYPE_INT64:
        assert_type(pyarrow.int64())
        lit = transaction_writer_pb2.Expression.Literal(int64=pa_value.as_py())
    elif column.offline_data_type.type == data_type_pb2.DataTypeEnum.DATA_TYPE_STRING:
        assert_type(pyarrow.string())
        lit = transaction_writer_pb2.Expression.Literal(str=pa_value.as_py())
    elif column.offline_data_type.type == data_type_pb2.DataTypeEnum.DATA_TYPE_TIMESTAMP:
        assert_type(pyarrow.timestamp("us", "UTC"))
        lit = transaction_writer_pb2.Expression.Literal()
        lit.timestamp.FromDatetime(pa_value.as_py())
    else:
        msg = f"Unsupported type {column.offline_data_type.type} in column {column.name}"
        raise Exception(msg)
    return transaction_writer_pb2.Expression(literal=lit)


def _binary_expr(
    op: transaction_writer_pb2.Expression.Binary.Op,
    left: transaction_writer_pb2.Expression,
    right: transaction_writer_pb2.Expression,
) -> transaction_writer_pb2.Expression:
    return transaction_writer_pb2.Expression(
        binary=transaction_writer_pb2.Expression.Binary(op=op, left=left, right=right)
    )


TRUE = transaction_writer_pb2.Expression(literal=transaction_writer_pb2.Expression.Literal(bool=True))


def _in_range(
    table: pyarrow.Table, column: schema_pb2.Column, end_inclusive: bool
) -> transaction_writer_pb2.Expression:
    """Returns a predicate Expression for values which are within the limits of the given column of the given limits table.

    :param table: The table
    :param: column: The column to test in this expression
    :param: Whether the predicate should include the end value
    """
    start_cond = _binary_expr(
        op=transaction_writer_pb2.Expression.Binary.OP_LE,
        left=_pyarrow_literal(table, column, row=0),
        right=transaction_writer_pb2.Expression(column=column),
    )
    end_cond = _binary_expr(
        op=transaction_writer_pb2.Expression.Binary.OP_LE
        if end_inclusive
        else transaction_writer_pb2.Expression.Binary.OP_LT,
        left=transaction_writer_pb2.Expression(column=column),
        right=_pyarrow_literal(table, column, row=1),
    )
    return _binary_expr(op=transaction_writer_pb2.Expression.Binary.OP_AND, left=start_cond, right=end_cond)


@dataclasses.dataclass
class DeltaWriter:
    def __init__(
        self,
        store_params: OfflineStoreParams,
        table_uri: str,
        join_keys: List[str],
        dynamodb_log_table_name: str,
        dynamodb_log_table_region: str,
        dynamodb_cross_account_role: Optional[AwsIamRole],
        kms_key_arn: str,
    ):
        self._feature_params = store_params
        self._table_uri = table_uri
        self._join_keys = join_keys
        self._fs, self._base_path = pyarrow.fs.FileSystem.from_uri(self._table_uri)
        self._adds: List[transaction_writer_pb2.AddFile] = []
        self._delete_uris: List[str] = []
        self._dynamodb_log_table_name = dynamodb_log_table_name
        self._dynamodb_log_table_region = dynamodb_log_table_region
        self._dynamodb_cross_account_role = dynamodb_cross_account_role
        self._current_transaction_writer: Optional[TransactionWriter] = None
        self._partitioning = pyarrow.dataset.partitioning(
            pyarrow.schema([(offline_store.TIME_PARTITION, pyarrow.string())]), flavor="hive"
        )
        self._kms_key_arn = kms_key_arn

    def _transaction_writer(self) -> TransactionWriter:
        if not self._current_transaction_writer:
            schema = self._feature_params.schema
            partition_column = schema_pb2.Column(
                name=offline_store.TIME_PARTITION,
                offline_data_type=data_type_pb2.DataType(type=data_type_pb2.DataTypeEnum.DATA_TYPE_STRING),
            )
            schema.columns.append(partition_column)
            init_args = transaction_writer_pb2.InitializeArgs(
                path=self._table_uri,
                id=self._feature_params.feature_view_id,
                name=self._feature_params.feature_view_name,
                description=f"Offline store for FeatureView {self._feature_params.feature_view_id} ({self._feature_params.feature_view_name})",
                schema=schema,
                partition_columns=[offline_store.TIME_PARTITION],
                dynamodb_log_table_name=self._dynamodb_log_table_name,
                dynamodb_log_table_region=self._dynamodb_log_table_region,
                kms_key_arn=self._kms_key_arn,
            )
            if self._dynamodb_cross_account_role is not None:
                init_args.cross_account_role_configs.dynamo_cross_account_role.CopyFrom(
                    self._dynamodb_cross_account_role
                )
            self._current_transaction_writer = TransactionWriter(init_args)
        return self._current_transaction_writer

    def _time_limits(self, time_interval: TimeInterval) -> pyarrow.Table:
        """Returns a Table specifying the limits of data affected by a materialization job.

        :param time_interval: The feature time interval
        :returns: A relation with one column for the timestamp key or anchor time, and one with the partition value
            corresponding to the first column. The first row will be the values for feature start time and the second for
            feature end time.
        """
        timestamp_key = self._feature_params.time_spec.timestamp_key
        timestamp_table = pyarrow.table({timestamp_key: [time_interval.start, time_interval.end]})

        if self._feature_params.batch_schedule is None:
            msg = "Batch schedule is required for batch materialization"
            raise Exception(msg)

        tree = AddTimePartitionNode(
            dialect=Dialect.DUCKDB,
            compute_mode=ComputeMode.RIFT,
            input_node=AddAnchorTimeNode(
                dialect=Dialect.DUCKDB,
                compute_mode=ComputeMode.RIFT,
                input_node=ConvertTimestampToUTCNode(
                    dialect=Dialect.DUCKDB,
                    compute_mode=ComputeMode.RIFT,
                    input_node=StagedTableScanNode(
                        dialect=Dialect.DUCKDB,
                        compute_mode=ComputeMode.RIFT,
                        staged_columns=(timestamp_key,),
                        staging_table_name="timestamp_table",
                    ).as_ref(),
                    timestamp_key=timestamp_key,
                ).as_ref(),
                feature_store_format_version=self._feature_params.feature_store_format_version,
                batch_schedule=self._feature_params.batch_schedule,
                timestamp_field=timestamp_key,
            ).as_ref(),
            time_spec=self._feature_params.time_spec,
        ).as_ref()
        conn = DuckDBContext.get_instance().get_connection()
        return conn.sql(tree.to_sql()).arrow()

    def _time_limit_predicate(self, interval: TimeInterval) -> transaction_writer_pb2.Expression:
        """Returns a predicate Expression matching offline store rows for materialization of the given interval."""
        table = self._time_limits(interval)
        time_spec = self._feature_params.time_spec
        time_column = next((col for col in self._feature_params.schema.columns if col.name == time_spec.time_column))
        partition_column = schema_pb2.Column(
            name=offline_store.TIME_PARTITION,
            offline_data_type=data_type_pb2.DataType(type=data_type_pb2.DataTypeEnum.DATA_TYPE_STRING),
        )
        predicate = _binary_expr(
            op=transaction_writer_pb2.Expression.Binary.OP_AND,
            left=_in_range(table, time_column, end_inclusive=False),
            right=_in_range(table, partition_column, end_inclusive=True),
        )
        return predicate

    def _filter_files(
        self,
        predicate: transaction_writer_pb2.Expression,
        filter_table: Callable[[pyarrow.dataset.Dataset], pyarrow.Table],
        force_overwrite: bool = False,
        **write_kwargs,
    ):
        paths = self._transaction_writer().read_for_update(predicate)
        deletes = []
        for path in paths:
            input_table = pyarrow.dataset.dataset(
                source=os.path.join(self._base_path, path),
                filesystem=self._fs,
                partitioning=self._partitioning,
            ).to_table()
            output_table = filter_table(input_table)
            if input_table.num_rows != output_table.num_rows or force_overwrite:
                deletes.append(path)
                if output_table.num_rows:
                    self.write(output_table, **write_kwargs)
        self._delete_uris.extend(deletes)

    def _filter_materialized_range_for_deletion(self, interval: TimeInterval) -> None:
        """Filters data within a materialized time range from parquet files in the offline store.

        :param interval: The feature data time interval to delete
        """
        time_spec = self._feature_params.time_spec
        conn = DuckDBContext.get_instance().get_connection()

        def table_filter(input_table: pyarrow.dataset.Dataset) -> pyarrow.Table:
            time_limit_table = self._time_limits(interval)
            # Add timezone to timestamps
            input_table = input_table.cast(patch_timestamps_in_arrow_schema(input_table.schema))
            # Not using pypika because it lacks support for ANTI JOIN
            return conn.sql(
                f"""
                WITH flattened_limits AS(
                    SELECT MIN("{time_spec.time_column}") AS start, MAX("{time_spec.time_column}") AS end
                    FROM time_limit_table
                )
                SELECT * FROM input_table
                LEFT JOIN flattened_limits
                ON input_table."{time_spec.time_column}" >= flattened_limits.start
                AND input_table."{time_spec.time_column}" < flattened_limits.end
                WHERE flattened_limits.start IS NULL
            """
            ).arrow()

        predicate = self._time_limit_predicate(interval)
        self._filter_files(predicate, table_filter)

    def transaction_exists(self, metadata: metadata_pb2.TectonDeltaMetadata) -> bool:
        """checks matching transaction metadata, which signals that a previous task attempt already wrote data
        If the task overwrites a previous materialization task interval then we treat it as a new transaction.
        # TODO (vitaly): replace with txnAppId since overwrite tasks might also have multiple attempts (redundant txns)

        :param metadata: transaction metadata
        :return: whether the same transaction has been executed before
        """
        return self._transaction_writer().has_commit_with_metadata(metadata)

    def delete_time_range(self, interval: TimeInterval) -> None:
        """Deletes previously materialized data within the interval if the interval overlaps with a previous task.

        High level process:
        1. Construct a Delta predicate expression matching the data we want to delete. This includes both a partition
           predicate to limit the files we have to look at, and a predicate on timestamp/anchor time which doesn't limit
           the files we have to consider, but can help with limit transaction conflicts.
        2. Mark files matching the predicate as read in the Delta transaction. This returns a list of files possibly
           matching the predicate.
        3. For each file:
           3a. Open it, filter out all data matching the predicate, and write out remaining data (if any) to a
               new file.
           3b. If any data was filtered out, add the old file to the list of deletes in the transaction. If any data remains,
               add the new file to the transaction.

        Implementation notes:
        1. This is racy if there is another job running at the same time. This behavior is the same as in Spark.
        2. We have corrected a bug that exists in Spark where we're not correctly selecting the data to delete: TEC-16681
        """
        print(f"Clearing prior data in range {interval.start} - {interval.end}")
        self._filter_materialized_range_for_deletion(interval)

    def write(
        self, table: Union[pyarrow.Table, pyarrow.RecordBatchReader], tags: Optional[Dict[str, str]] = None
    ) -> List[str]:
        """Writes a pyarrow Table to the Delta table at base_uri partitioned by the TIME_PARTITION column.

        Returns a list of URIs for the written file(s).

        This does NOT commit to the Delta log. Call commit() after calling this to commit your changes.
        """

        adds = []
        failed = False

        def visit_file(f: WrittenFile):
            try:
                path = f.path
                _, prefix, relative = path.partition(self._base_path)
                assert prefix == self._base_path, f"Written path is not relative to base path: {path}"
                path_pieces = relative.split("/")
                partition = path_pieces[1]
                k, eq, v = partition.partition("=")
                assert k == offline_store.TIME_PARTITION and eq == "=", f"Unexpected partition format: {path}"
                stats = get_file_stats_from_metadata(
                    f.metadata,
                    num_indexed_cols=-1,  # since we specify columns_to_collect_stats this should be -1
                    columns_to_collect_stats=self._join_keys,
                )
                serialized_stats = json.dumps(stats, cls=DeltaJSONEncoder)
                # somehow Delta Standalone wants this to be double-serialized with all special JSON literals escaped
                serialized_stats = json.dumps(serialized_stats)
                adds.append(
                    transaction_writer_pb2.AddFile(
                        uri=self._table_uri + relative,
                        partition_values={k: v},
                        tags=tags,
                        stats=serialized_stats,
                    )
                )
            except Exception as e:
                # Pyarrow logs and swallows exceptions from this function, so we need some other way of knowing there
                # was a # failure
                nonlocal failed
                failed = True
                raise e

        pyarrow.dataset.write_dataset(
            data=table,
            filesystem=self._fs,
            base_dir=self._base_path,
            format=pyarrow.dataset.ParquetFileFormat(),
            file_options=PARQUET_WRITE_OPTIONS,
            basename_template=f"{uuid.uuid4()}-part-{{i}}.parquet",
            partitioning=self._partitioning,
            file_visitor=visit_file,
            existing_data_behavior="overwrite_or_ignore",
            max_partitions=365 * 100,
            max_rows_per_file=PARQUET_MAX_ROWS_PER_FILE,
        )

        if failed:
            msg = "file visitor failed"
            raise Exception(msg)

        self._adds.extend(adds)
        return [add.uri for add in adds]

    def delete_keys(self, keys: pyarrow.Table):
        def filter_table(input_table: pyarrow.dataset.Dataset) -> pyarrow.Table:
            conn = DuckDBContext.get_instance().get_connection()
            return conn.sql(
                f"""
                SELECT * FROM input_table
                ANTI JOIN keys
                USING ({", ".join(keys.column_names)})
                """
            ).arrow()

        # It's necessary to scan the entire table anyway, so we just use True as the predicate.
        #
        # In theory this might be missing out on some optimizations to avoid conflicting queries, but in
        # practice the only types of conflicts we could avoid would be key deletion operations on
        # disjoint sets of keys and also end up only touching disjoint sets of files. This is probably not very likely
        # to occur.
        return self._filter_files(TRUE, filter_table)

    def run_z_order_optimization(self):
        partitions = self._transaction_writer().get_partitions_not_marked_with({Z_ORDER_TAG: Z_ORDER_TAG_VALUE})
        partition_column = schema_pb2.Column(
            name=offline_store.TIME_PARTITION,
            offline_data_type=data_type_pb2.DataType(type=data_type_pb2.DataTypeEnum.DATA_TYPE_STRING),
        )
        conn = DuckDBContext.get_instance().get_connection()

        def table_sorted(input_table: pyarrow.dataset.Dataset) -> pyarrow.Table:
            return conn.from_arrow(input_table).order(", ".join(self._join_keys)).arrow()

        for partition in partitions:
            partition_predicate = _binary_expr(
                op=transaction_writer_pb2.Expression.Binary.OP_EQ,
                left=transaction_writer_pb2.Expression(column=partition_column),
                right=transaction_writer_pb2.Expression(
                    literal=transaction_writer_pb2.Expression.Literal(str=partition[offline_store.TIME_PARTITION])
                ),
            )
            self._filter_files(
                partition_predicate, table_sorted, force_overwrite=True, tags={Z_ORDER_TAG: Z_ORDER_TAG_VALUE}
            )

    def commit(self, metadata: Optional[metadata_pb2.TectonDeltaMetadata] = None) -> Optional[int]:
        """Returns version of commit if it was successful"""
        if not self._adds and not self._delete_uris:
            # nothing to commit
            return

        args = transaction_writer_pb2.UpdateArgs(
            add_files=self._adds, delete_uris=self._delete_uris, user_metadata=metadata
        )
        try:
            return self._transaction_writer().update(args).committed_version
        except DeltaConcurrentModificationException:
            # Commit should be retried together with new write.
            self.abort()

            raise
        finally:
            self._reset_state()

    def transaction(self, metadata: Optional[TectonDeltaMetadata] = None) -> Callable[[TxnFn], TxnFn]:
        """Returns a decorator which wraps a function in a Delta transaction.

        If the function returns successfully, the Delta transaction will be committed automatically. Any exceptions will
        cause an aborted transaction.

        Any Delta conflicts which occur will result in the function being retried in a new transaction.

        :param metadata: Optional metadata to be added to the transaction.
        """

        def decorator(f: TxnFn, max_attempts=5) -> TxnFn:
            @functools.wraps(f)
            def wrapper() -> R:
                for attempt in range(1, max_attempts + 1):
                    r = f()
                    try:
                        self.commit(metadata)
                    except DeltaConcurrentModificationException:
                        if attempt >= max_attempts:
                            raise
                        print(f"Delta commit attempt {attempt} failed. Retrying...")
                    finally:
                        self.abort()
                    return r

            return wrapper

        return decorator

    def abort(self):
        """
        Abort the transaction by cleaning up any files and state.
        Clean up created parquet files that were not part of a successful commit.
        """
        for add_file in self._adds:
            self._fs.delete_file(path_from_uri(add_file.uri))
        self._reset_state()

    def _reset_state(self):
        self._current_transaction_writer = None
        self._adds = []
        self._delete_uris = []


def path_from_uri(uri):
    parts = urlparse(uri)
    return parts.netloc + parts.path



import tempfile
import time
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import pytest
import requests

from tecton.identities import okta
from tecton_core import conf


def fake_get_tokens_helper(endpoint, params):
    msg = "simulating refresh token expired"
    raise requests.RequestException(msg)


class OktaTest(TestCase):
    def test_refresh_token_success(self):
        class FakeRefreshFlow:
            def get_tokens(*args):
                return "NEW_ACCESS_TOKEN", 0, 0, 0

        # Test local Tecton config file
        with tempfile.NamedTemporaryFile() as f:
            f.write(
                (
                    '{"OAUTH_ACCESS_TOKEN": "dummy", "OAUTH_REFRESH_TOKEN": "dummy", "OAUTH_ACCESS_TOKEN_EXPIRATION": "%s"}'
                    % (str(time.time() - 60))
                ).encode()
            )
            f.flush()
            with patch("tecton.conf._LOCAL_TECTON_TOKENS_FILE", Path(f.name)):
                conf._init_configs()
                with patch("tecton.identities.okta.RefreshFlow", FakeRefreshFlow):
                    assert okta.get_token_refresh_if_needed() == "NEW_ACCESS_TOKEN"

    def test_refresh_token_failure(self):
        conf.set_okta_tokens(access_token="dummy", access_token_expiration=time.time() - 60, refresh_token="dummy")
        with patch("tecton.identities.okta.get_tokens_helper", fake_get_tokens_helper):
            with pytest.raises(PermissionError, match="Authorization token expired"):
                okta.get_token_refresh_if_needed()

    def test_get_user_profile_refresh_token_expired(self):
        conf.set_okta_tokens(access_token="dummy", access_token_expiration=time.time() - 60, refresh_token="dummy")
        with patch("tecton.identities.okta.get_tokens_helper", fake_get_tokens_helper):
            with pytest.raises(PermissionError, match="Authorization token expired"):
                okta.get_user_profile()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protoc-gen-openapiv2/options/annotations.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
from protoc_gen_openapiv2.options import openapiv2_pb2 as protoc__gen__openapiv2_dot_options_dot_openapiv2__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.protoc-gen-openapiv2/options/annotations.proto\x12)grpc.gateway.protoc_gen_openapiv2.options\x1a google/protobuf/descriptor.proto\x1a,protoc-gen-openapiv2/options/openapiv2.proto:~\n\x11openapiv2_swagger\x12\x1c.google.protobuf.FileOptions\x18\x92\x08 \x01(\x0b\x32\x32.grpc.gateway.protoc_gen_openapiv2.options.SwaggerR\x10openapiv2Swagger:\x86\x01\n\x13openapiv2_operation\x12\x1e.google.protobuf.MethodOptions\x18\x92\x08 \x01(\x0b\x32\x34.grpc.gateway.protoc_gen_openapiv2.options.OperationR\x12openapiv2Operation:~\n\x10openapiv2_schema\x12\x1f.google.protobuf.MessageOptions\x18\x92\x08 \x01(\x0b\x32\x31.grpc.gateway.protoc_gen_openapiv2.options.SchemaR\x0fopenapiv2Schema:u\n\ropenapiv2_tag\x12\x1f.google.protobuf.ServiceOptions\x18\x92\x08 \x01(\x0b\x32..grpc.gateway.protoc_gen_openapiv2.options.TagR\x0copenapiv2Tag:~\n\x0fopenapiv2_field\x12\x1d.google.protobuf.FieldOptions\x18\x92\x08 \x01(\x0b\x32\x35.grpc.gateway.protoc_gen_openapiv2.options.JSONSchemaR\x0eopenapiv2FieldBHZFgithub.com/grpc-ecosystem/grpc-gateway/v2/protoc-gen-openapiv2/optionsb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protoc_gen_openapiv2.options.annotations_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:
  google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(openapiv2_swagger)
  google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(openapiv2_operation)
  google_dot_protobuf_dot_descriptor__pb2.MessageOptions.RegisterExtension(openapiv2_schema)
  google_dot_protobuf_dot_descriptor__pb2.ServiceOptions.RegisterExtension(openapiv2_tag)
  google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(openapiv2_field)

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'ZFgithub.com/grpc-ecosystem/grpc-gateway/v2/protoc-gen-openapiv2/options'
# @@protoc_insertion_point(module_scope)



"""A multi-producer, multi-consumer closable queue."""

import threading
from collections import deque
from time import monotonic as time


class Empty(Exception):
    "Exception raised by Queue.get(block=0)/get_nowait()."


class Full(Exception):
    "Exception raised by Queue.put(block=0)/put_nowait()."


class Closed(Exception):
    """Exception to indicate that the ClosableQueue is closed."""



class ClosableQueue:
    """Create a queue object with a given maximum size.

    If maxsize is <= 0, the queue size is infinite.

    This adds an extra behavior where a producer can `close` the queue,
    to indicate that no new items will be added. Semantics:
      `get`: When a queue is closed, `get` calls will succeed until the queue is empty,
             at which point further get calls will raise `Closed`. All pending `get`,
             which happens when the queue is already empty, will be pre-empted and raise `Closed`.

      `put`: When a queue is closed, `put` calls raise `Closed`. All pending `put`
             will be pre-empted and also raise `Closed`.

    NOTE: This implementation is inspired by the stdlib implementation.
    The code has been adapted to handle the semantics of the `close`.
    Reference: https://github.com/python/cpython/blob/3.8/Lib/queue.py
    """

    def __init__(self, maxsize=0):
        self.maxsize = maxsize
        self._init(maxsize)

        # mutex must be held whenever the queue is mutating.  All methods
        # that acquire mutex must release it before returning.  mutex
        # is shared between the three conditions, so acquiring and
        # releasing the conditions also acquires and releases mutex.
        self.mutex = threading.Lock()

        # Notify not_empty whenever an item is added to the queue; a
        # thread waiting to get is notified then.
        self.not_empty = threading.Condition(self.mutex)

        # Notify not_full whenever an item is removed from the queue;
        # a thread waiting to put is notified then.
        self.not_full = threading.Condition(self.mutex)

        # Notify all_tasks_done whenever the number of unfinished tasks
        # drops to zero; thread waiting to join() is notified to resume
        self.all_tasks_done = threading.Condition(self.mutex)
        self.unfinished_tasks = 0

        self._is_closed = False

    def task_done(self):
        """Indicate that a formerly enqueued task is complete.

        Used by Queue consumer threads.  For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items
        have been processed (meaning that a task_done() call was received
        for every item that had been put() into the queue).

        Raises a ValueError if called more times than there were items
        placed in the queue.
        """
        with self.all_tasks_done:
            unfinished = self.unfinished_tasks - 1
            if unfinished <= 0:
                if unfinished < 0:
                    msg = "task_done() called too many times"
                    raise ValueError(msg)
                self.all_tasks_done.notify_all()
            self.unfinished_tasks = unfinished

    def join(self):
        """Blocks until all items in the Queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer thread calls task_done()
        to indicate the item was retrieved and all work on it is complete.

        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        with self.all_tasks_done:
            while self.unfinished_tasks:
                self.all_tasks_done.wait()

    def qsize(self):
        """Return the approximate size of the queue (not reliable!)."""
        with self.mutex:
            return self._qsize()

    def put(self, item, block=True, timeout=None):
        """Put an item into the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until a free slot is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Full exception if no free slot was available within that time.
        Otherwise ('block' is false), put an item on the queue if a free slot
        is immediately available, else raise the Full exception ('timeout'
        is ignored in that case).
        """
        with self.not_full:
            if self._is_closed:
                msg = "Cannot add to a closed queue."
                raise Closed(msg)

            if self.maxsize > 0:
                if not block:
                    if self._qsize() >= self.maxsize:
                        raise Full
                elif timeout is None:
                    while self._qsize() >= self.maxsize and not self._is_closed:
                        self.not_full.wait()
                elif timeout < 0:
                    msg = "'timeout' must be a non-negative number"
                    raise ValueError(msg)
                else:
                    endtime = time() + timeout
                    while self._qsize() >= self.maxsize and not self._is_closed:
                        remaining = endtime - time()
                        if remaining <= 0.0:
                            raise Full
                        self.not_full.wait(remaining)

            # Handle the case where we are done waiting because the
            # queue is closed.
            if self._is_closed:
                msg = "Cannot add to a closed queue."
                raise Closed(msg)

            self._put(item)
            self.unfinished_tasks += 1
            self.not_empty.notify()

    def get(self, block=True, timeout=None):
        """Remove and return an item from the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        """
        with self.not_empty:
            # Handle the case where the queue is closed and already empty.
            # We should not have to wait.
            if not self._qsize() and self._is_closed:
                msg = "Cannot get from a closed queue with no items."
                raise Closed(msg)

            if not block:
                if not self._qsize():
                    raise Empty
            elif timeout is None:
                while not self._qsize() and not self._is_closed:
                    self.not_empty.wait()
            elif timeout < 0:
                msg = "'timeout' must be a non-negative number"
                raise ValueError(msg)
            else:
                endtime = time() + timeout
                while not self._qsize() and not self._is_closed:
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise Empty
                    self.not_empty.wait(remaining)

            # Handle the case where we are done waiting because the
            # queue is closed.
            if self._is_closed and not self._qsize():
                msg = "Cannot get from a closed queue with no items."
                raise Closed(msg)

            item = self._get()
            self.not_full.notify()
            return item

    def put_nowait(self, item):
        """Put an item into the queue without blocking.

        Only enqueue the item if a free slot is immediately available.
        Otherwise raise the Full exception.
        """
        return self.put(item, block=False)

    def get_nowait(self):
        """Remove and return an item from the queue without blocking.

        Only get an item if one is immediately available. Otherwise
        raise the Empty exception.
        """
        return self.get(block=False)

    def close(self) -> None:
        with self.mutex:
            if self._is_closed:
                msg = "Cannot `close` an already closed queue."
                raise Closed(msg)
            self._is_closed = True

        # We need to notify waiters that this queue is closed.
        with self.not_empty:
            self.not_empty.notify_all()

        # We need to notify waiters that this queue is closed.
        with self.not_full:
            self.not_full.notify_all()

    def is_closed(self) -> bool:
        with self.mutex:
            return self._is_closed

    # Override these methods to implement other queue organizations
    # (e.g. stack or priority queue).
    # These will only be called with appropriate locks held

    # Initialize the queue representation
    def _init(self, maxsize):
        self.queue = deque()

    def _qsize(self):
        return len(self.queue)

    # Put a new item in the queue
    def _put(self, item):
        self.queue.append(item)

    # Get an item from the queue
    def _get(self):
        return self.queue.popleft()



# This is Spark specific job_metadata module. See common.job_metadata for common functions

import logging
import time
from typing import Optional

from tecton_core.id_helper import IdHelper
from tecton_materialization.common.job_metadata import IDEMPOTENCE_KEY_ATTRIBUTE
from tecton_materialization.common.job_metadata import LAST_UPDATED_ATTRIBUTE
from tecton_materialization.common.job_metadata import RUN_ID_PREFIX
from tecton_materialization.common.job_metadata import TTL_ATTRIBUTE
from tecton_materialization.common.job_metadata import TTL_DURATION_SECONDS
from tecton_materialization.common.job_metadata import VALUE_ATTRIBUTE
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_materialization.materialization_utils import get_statsd_client
from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadata
from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadataTableType


__all__ = (
    "write_checkpoint",
    "is_checkpoint_complete",
    "check_spark_job_uniqueness",
)

try:
    import boto3
    from botocore.errorfactory import ClientError

    from tecton_materialization.common.job_metadata_aws import _dynamodb_client
except ImportError:
    # not available and unused in dataproc
    boto3 = None
    ClientError = None
    _dynamodb_client = None

logger = logging.getLogger(__name__)


# TODO(Alex): Migrate checkpointing into JMT
# We use task as the idempotence key for checkpointing because we don't need to recompute the same tiles already processed within the same task.
# In a different task, they may correspond to a different store type, or be an overwrite, so we'd want to still process them.
# The run_id isn't required but could be useful for debugging.
def write_checkpoint(spark, materialization_task_params, anchor_time, run_id):
    dynamodb = _dynamodb_client(materialization_task_params)
    table = materialization_task_params.spark_job_execution_table
    idempotence_key = materialization_task_params.materialization_task_id + "@" + str(anchor_time)
    now_seconds = int(time.time())
    statsd_client = get_statsd_client(spark)
    dynamodb.put_item(
        TableName=table,
        Item={
            IDEMPOTENCE_KEY_ATTRIBUTE: {"S": idempotence_key},
            VALUE_ATTRIBUTE: {"S": f"{RUN_ID_PREFIX}{run_id}"},
            TTL_ATTRIBUTE: {"N": str(now_seconds + TTL_DURATION_SECONDS)},
            LAST_UPDATED_ATTRIBUTE: {"N": str(now_seconds)},
        },
    )
    statsd_client.incr("materialization.checkpoint_write", 1)


def is_checkpoint_complete(spark, materialization_task_params, anchor_time):
    dynamodb = _dynamodb_client(materialization_task_params)
    table = materialization_task_params.spark_job_execution_table
    idempotence_key = materialization_task_params.materialization_task_id + "@" + str(anchor_time)
    now_seconds = int(time.time())
    statsd_client = get_statsd_client(spark)
    try:
        dynamodb.put_item(
            TableName=table,
            Item={
                IDEMPOTENCE_KEY_ATTRIBUTE: {"S": idempotence_key},
                VALUE_ATTRIBUTE: {"S": ""},
                TTL_ATTRIBUTE: {"N": str(now_seconds + TTL_DURATION_SECONDS)},
                LAST_UPDATED_ATTRIBUTE: {"N": str(now_seconds)},
            },
            ConditionExpression=f"attribute_not_exists({IDEMPOTENCE_KEY_ATTRIBUTE}) OR #val = :val",
            ExpressionAttributeNames={"#val": VALUE_ATTRIBUTE},
            ExpressionAttributeValues={":val": {"S": ""}},
        )
        statsd_client.incr("materialization.checkpoint_check_incomplete", 1)
        return False
    except ClientError as e:
        # Condition failed means we've previously committed the checkpoint
        if e.response["Error"]["Code"] == "ConditionalCheckFailedException":
            statsd_client.incr("materialization.checkpoint_check_complete", 1)
            return True
        else:
            raise e


def check_spark_job_uniqueness(
    materialization_task_params,
    run_id,
    spark,
    step,
    skip_legacy_execution_table_check,
    job_metadata_client: JobMetadataClient,
):
    if step not in (1, None):
        return
    if (
        not skip_legacy_execution_table_check
        and materialization_task_params.job_metadata_table_type != JobMetadataTableType.JOB_METADATA_TABLE_TYPE_GCS
    ):
        dynamodb = _dynamodb_client(materialization_task_params)
        table = materialization_task_params.spark_job_execution_table
        idempotence_key = materialization_task_params.idempotence_key

        # TODO: remove once job_metadata_table is enabled everywhere
        statsd_client = get_statsd_client(spark)
        try:
            existing_record = dynamodb.get_item(
                TableName=table,
                Key={IDEMPOTENCE_KEY_ATTRIBUTE: {"S": idempotence_key}},
                ConsistentRead=True,
            ).get("Item", None)
            statsd_client.incr("materialization.dynamo_get_item_success", 1)
        except ClientError:
            existing_record = None
            statsd_client.incr("materialization.dynamo_get_item_errors", 1)

        try:
            now_seconds = int(time.time())
            dynamodb.put_item(
                TableName=table,
                Item={
                    IDEMPOTENCE_KEY_ATTRIBUTE: {"S": idempotence_key},
                    VALUE_ATTRIBUTE: {"S": f"{RUN_ID_PREFIX}{run_id}"},
                    TTL_ATTRIBUTE: {"N": str(now_seconds + TTL_DURATION_SECONDS)},
                    LAST_UPDATED_ATTRIBUTE: {"N": str(now_seconds)},
                },
                ConditionExpression=f"attribute_not_exists({IDEMPOTENCE_KEY_ATTRIBUTE})",
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "ConditionalCheckFailedException":
                if existing_record:
                    logger.error(f"Existing Spark unique key record found: {existing_record}")
                else:
                    logger.error(f"Conditional check failed, but no record found for {idempotence_key}")
                msg = f"Value is already set for idempotence_key: {idempotence_key}"
                raise RuntimeError(msg) from e
            raise e

    # TODO: remove above code and remove jobidempotencemanager code once NEW_JOB_EXECUTION_TABLE migration is done
    # after all customers have dynamodb:GetItem permission
    def updater(job_metadata: JobMetadata) -> Optional[JobMetadata]:
        new_proto = JobMetadata()
        new_proto.CopyFrom(job_metadata)
        if new_proto.spark_execution_info.HasField("run_id"):
            attempt_id = IdHelper.to_string(materialization_task_params.attempt_id)
            msg = f"Value is already set for attempt_id: {attempt_id}: {new_proto.spark_execution_info.run_id}"
            raise RuntimeError(msg)
        elif new_proto.spark_execution_info.is_revoked:
            msg = "Job cancelled by orchestrator before lock acquired"
            raise RuntimeError(msg)
        new_proto.spark_execution_info.run_id = run_id
        return new_proto

    if materialization_task_params.HasField("job_metadata_table"):
        job_metadata_client.update(updater)



import abc
import datetime
from types import MappingProxyType
from typing import List
from typing import Optional
from typing import Union

from tecton._internals.tecton_pydantic import StrictModel
from tecton._internals.tecton_pydantic import pydantic_v1
from tecton.aggregation_functions import AggregationFunction
from tecton.framework.configs import LifetimeWindow
from tecton.framework.configs import TimeWindow
from tecton.framework.configs import TimeWindowSeries
from tecton.framework.configs import build_aggregation_proto
from tecton.types import Field
from tecton.types import SdkDataType
from tecton.types import String
from tecton_proto.args import feature_view__client_pb2 as feature_view_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2


class Feature(StrictModel, abc.ABC):
    column: str
    column_dtype: SdkDataType

    @abc.abstractmethod
    def _to_proto(self, **kwargs):
        pass


class Aggregate(Feature):
    """
    The `Aggregate` class describes an aggregation feature that is applied to a Batch or Stream Feature View via `features` param.

    :param column: Column name of the feature.
    :type column: str
    :param column_dtype: The data type of the column.
    :type column_dtype: SdkDataType
    :param function: One of the built-in aggregation functions, such as "sum", "count", `last(2)` etc.
    :type function: Union[str, AggregationFunction]
    :param time_window: The window_size and optional offset over which to aggregate over.
    :type time_window: Union[TimeWindow, TimeWindowSeries, LifetimeWindow, datetime.timedelta]
    :param name: The name of this feature. Defaults to an autogenerated name, e.g. transaction_count_7d_1d.
    :type name: str

    .. code-block:: python
        from tecton import Aggregate, batch_feature_view, TimeWindow
        from tecton.types import Int64
        from datetime import timedelta


        @batch_feature_view(
            # ...
            features=[
                Aggregate(
                    column="my_column",
                    column_dtype=Int64,
                    function="mean",
                    time_window=TimeWindow(window_size=timedelta(days=7)),
                ),
                Aggregate(
                    column="another_column",
                    column_dtype=Int64,
                    function="mean",
                    time_window=TimeWindow(window_size=timedelta(days=1)),
                    name="1d_average",
                ),
            ],
        )
        def my_fv(data_source):
            pass

    """

    function: AggregationFunction
    time_window: Union[TimeWindow, TimeWindowSeries, LifetimeWindow]
    name: Optional[str] = None

    def __init__(
        self,
        column: str,
        column_dtype: SdkDataType,
        function: Union[str, AggregationFunction],
        time_window: Union[TimeWindow, TimeWindowSeries, LifetimeWindow, datetime.timedelta],
        name: Optional[str] = None,
    ):
        super().__init__(
            column=column, column_dtype=column_dtype, function=function, time_window=time_window, name=name
        )

    @pydantic_v1.validator("time_window", pre=True)
    def timedelta_to_time_window(cls, v):
        if isinstance(v, datetime.timedelta):
            return TimeWindow(window_size=v)
        return v

    @pydantic_v1.validator("function", pre=True)
    def str_to_aggregation_function(cls, v):
        if isinstance(v, str):
            return AggregationFunction(base_name=v, resolved_name=v, params=MappingProxyType({}))
        return v

    def _to_proto(
        self,
        aggregation_interval: datetime.timedelta,
        is_continuous: bool,
        compaction_enabled: bool = False,
        is_streaming_fv: bool = False,
    ) -> feature_view_pb2.FeatureAggregation:
        proto = build_aggregation_proto(
            self.name,
            self.column,
            self.function,
            self.time_window,
            aggregation_interval,
            is_continuous,
            compaction_enabled,
            is_streaming_fv,
        )
        proto.column_dtype.CopyFrom(self.column_dtype.tecton_type.proto)
        return proto


class Attribute(Feature):
    """
    The `Attribute` class describes an attribute feature that is applied to a Feature View or Feature Table via `features` param.

    :param column: Column name of the feature.
    :type column: str
    :param column_dtype: The data type of the column.
    :type column_dtype: SdkDataType


    .. code-block:: python
    from tecton import Attribute, batch_feature_view
    from tecton.types import String

    @batch_feature_view(
        # ...
        features=[Attribute(column="my_column", column_dtype=String)],
    )
    def my_fv(data_source):
        pass
    """

    def __init__(self, column: str, column_dtype: SdkDataType):
        super().__init__(column=column, column_dtype=column_dtype)

    def _to_proto(self):
        return feature_view_pb2.Attribute(
            column=self.column,
            column_dtype=self.column_dtype.tecton_type.proto,
        )


class Embedding(Feature):
    """
    The `Embedding` class describes an embedding feature that is applied to a Batch Feature View via `features` param.

    :param column: The name of the column to be embedded.
    :type column: str
    :param column_dtype: The data type of the column to be embedded.
    :type column_dtype: SdkDataType
    :param model: The model name that is used to compute the embedding feature. Check https://docs.tecton.ai/docs/beta/defining-features/feature-views/embeddings for all supported models.
    :type model: str
    :param name: The name of this feature. Defaults to an autogenerated name, e.g. "my_embedding_feature".
    :type name: str


    .. code-block:: python
    from tecton import Embedding, batch_feature_view
    from tecton.types import String

    @batch_feature_view(
        # ...
        features=[
            Embedding(
                column="my_column",
                column_dtype=String,
                model="sentence-transformers/all-MiniLM-L6-v2",
                name="my_embedding_feature")
        ],
    )
    def my_fv(data_source):
        return data_source
    """

    model: str
    name: Optional[str] = None

    def __init__(self, column: str, column_dtype: SdkDataType, model: str, name: Optional[str] = None):
        super().__init__(column=column, column_dtype=column_dtype, model=model, name=name)

    def _to_proto(self) -> feature_view_pb2.Embedding:
        name = self.name if self.name else f"{self.column}_embedding"
        return feature_view_pb2.Embedding(
            name=name, column=self.column, column_dtype=self.column_dtype.tecton_type.proto, model=self.model
        )


class Inference(Feature):
    """
    The `Inference` class describes a model inference feature that is applied to a Batch Feature View via `features` param.

    :param input_columns: A list of columns that are used as inputs to the model.
    :type input_columns: List[Field]
    :param model: The model name that is used to compute the inference feature. The model needs to be registered in advance using `tecton model` CLI.
    :type model: str
    :param name: The name of this feature. Defaults to an autogenerated name, e.g. "my_inference_feature".
    :type name: Optional[str]


    .. code-block:: python
    from tecton import Inference, batch_feature_view
    from tecton.types import String, Int64

    @batch_feature_view(
        # ...
        features=[
            Inference(
                input_columns=[Field("my_field1", String), Field("my_field2", Int64)],
                model="my_custom_model",
                name="my_inference_feature")
        ],
    )
    def my_fv(data_source):
        return data_source
    """

    model: str
    input_columns: List[Field]
    name: Optional[str]

    def __init__(self, model: str, input_columns: List[Field], name: str):
        # TODO(jiadong): `column` and `column_dtype` are required right now. Need to refactor `Features` to accommodate new `Inferece` class.
        super().__init__(model=model, input_columns=input_columns, name=name, column="", column_dtype=String)

    def _to_proto(self) -> feature_view_pb2.Inference:
        # TODO(jiadong): Figure out a default name for `Inference` feature.
        return feature_view_pb2.Inference(
            model=self.model,
            input_columns=[
                schema_pb2.Column(name=field.name, offline_data_type=field.dtype.tecton_type.proto)
                for field in self.input_columns
            ],
            name=self.name,
        )



from typing import Optional

from tecton.identities import okta
from tecton_core import conf
from tecton_core.metadata_service_impl.providers import AuthProvider


class InternalAuthProvider(AuthProvider):
    def get_auth_header(self) -> Optional[str]:
        token = okta.get_token_refresh_if_needed()
        if token:
            return f"Bearer {token}"

        token = conf.get_or_none("TECTON_API_KEY")
        if token:
            return f"Tecton-key {token}"

        return None



import importlib.resources
import importlib.util
import sys
from pathlib import Path


def find_spark():
    spec = importlib.util.find_spec("pyspark")
    if spec:
        # pyspark exists in the system path
        return
    tecton_spec = importlib.util.find_spec("tecton")
    vendor_path = Path(tecton_spec.origin).parent / "vendor/pyspark"
    if not vendor_path.exists():
        msg = f"Could not locate system pyspark or vendored pyspark (looked in {vendor_path})"
        raise Exception(msg)
    # insert at the beginning. This is important because although we already know pyspark isn't anywhere in the path,
    # we've also vendored the corresponding version of py4j
    sys.path.insert(0, str(vendor_path))
    if importlib.util.find_spec("pyspark") is None:
        msg = f"Could not import vendored pyspark (added path {vendor_path})"
        raise Exception(msg)



from __future__ import annotations

import dataclasses
import gc
import sys
import traceback
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import Union

import pyarrow
import pyarrow.parquet
import torch.cuda


if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

if sys.version_info >= (3, 10):
    from typing import Concatenate
    from typing import ParamSpec
else:
    from typing_extensions import Concatenate
    from typing_extensions import ParamSpec


# NOTE [ Python Traceback Reference Cycle Problem ]
#
# When using sys.exc_info(), it is important to **not** store the exc_info[2],
# which is the traceback, because otherwise you will run into the traceback
# reference cycle problem, i.e., the traceback holding reference to the frame,
# and the frame (which holds reference to all the object in its temporary scope)
# holding reference the traceback.
#  https://github.com/python/cpython/issues/75188


class KeyErrorMessage(str):
    r"""str subclass that returns itself in repr"""

    def __repr__(self):
        return self


class ExceptionWrapper:
    """Wraps an exception and traceback for exception handling in a multi-threaded execution setting."""

    def __init__(self, exc_info=None, where="in background"):
        # It is important that we don't store exc_info, see
        # NOTE [ Python Traceback Reference Cycle Problem ]
        if exc_info is None:
            exc_info = sys.exc_info()
        self.exc_type = exc_info[0]
        self.exc_msg = "".join(traceback.format_exception(*exc_info))
        self.where = where

    def reraise(self):
        r"""Reraises the wrapped exception in the current thread"""
        # Format a message such as: "Caught ValueError in DataLoader worker
        # process 2. Original Traceback:", followed by the traceback.
        msg = f"Caught {self.exc_type.__name__} {self.where}.\nOriginal {self.exc_msg}"
        if self.exc_type == KeyError:
            # KeyError calls repr() on its argument (usually a dict key). This
            # makes stack traces unreadable. It will not be changed in Python
            # (https://github.com/python/cpython/issues/46903), so we work around it.
            msg = KeyErrorMessage(msg)
        elif getattr(self.exc_type, "message", None):
            # Some exceptions have first argument as non-str but explicitly
            # have message field
            raise self.exc_type(message=msg)
        try:
            exception = self.exc_type(msg)
        except TypeError:
            # If the exception takes multiple arguments, don't try to
            # instantiate since we don't know how to
            raise RuntimeError(msg) from None
        raise exception


Item = Union[pyarrow.RecordBatch, ExceptionWrapper]
P = ParamSpec("P")
PreprocessorCallable = Callable[Concatenate[Item, P], Iterable[Item]]
ModelCallable = Callable[Concatenate[Item, P], Item]


def data_preprocessor_one_step(item: Item, func: PreprocessorCallable, func_kwargs: P.kwargs) -> Iterable[Item]:
    if isinstance(item, ExceptionWrapper):
        return (item,)

    try:
        return func(item, **func_kwargs)
    except Exception:
        return (ExceptionWrapper(where="in data preprocessor"),)


def model_inference_one_step(item: Item, func: ModelCallable, func_kwargs: P.kwargs) -> Item:
    if isinstance(item, ExceptionWrapper):
        return item

    try:
        return func(item, **func_kwargs)
    except Exception:
        return ExceptionWrapper(where="in model inference")


class FuncConfig(Protocol):
    def load(self) -> FuncConfig: ...

    def kwargs(self) -> Mapping[str, Any]: ...


def gc_all(device: Optional[Union[str, int]] = None) -> None:
    """
    Run the garbage collector for the main thread and the GPU
    """
    # Python GC before cleaning up GPU memory to ensure that we release
    # unreferenced PyTorch tensors.
    gc.collect()
    if torch.cuda.is_available():
        if device is not None:
            with torch.cuda.device(device):
                torch.cuda.empty_cache()


@dataclasses.dataclass
class CudaMemoryUsageData:
    # Memory allocated before measurements started
    starting_alloc_bytes: int
    # Peak allocated memory during the measurement period
    max_alloc_bytes: int
    # Memory allocated during the measurement period. max_alloc - starting_alloc
    new_alloc_bytes: int
    # Device memory - starting_alloc
    starting_free_memory_bytes: int
    # The total memory on the device
    device_memory_bytes: int
    # Map of memory stats provided by torch.cuda.memory_stats
    _memory_stats: Mapping[str, Any]

    def max_token_budget_from_mem_alloc(self, sample_token_budget: int) -> int:
        """
        1. Calculates what proportion of available memory was allocated during the execution
        2. Calculates how of those "sample batches" would fit into available memory
        3. Multiplies the token budget of the sample batch by how many batches would've fit into memory
        """
        proportion_of_free_memory_allocated = self.new_alloc_bytes / self.starting_free_memory_bytes
        token_budget_multiplier = 1 / proportion_of_free_memory_allocated
        return int(sample_token_budget * token_budget_multiplier)


class CudaMemoryTracker:
    def __init__(self, device: Optional[str] = None) -> None:
        self.device = device if device is not None else torch.cuda.current_device()
        self._data: Optional[CudaMemoryUsageData] = None
        self._device_memory = None
        self._starting_alloc = None
        self._starting_free_memory = None

    def __enter__(self) -> "CudaMemoryTracker":
        gc_all(self.device)
        torch.cuda.reset_peak_memory_stats(self.device)
        self._device_memory = torch.cuda.get_device_properties(self.device).total_memory
        self._starting_alloc = torch.cuda.memory_allocated(self.device)
        self._starting_free_memory = self._device_memory - self._starting_alloc
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        torch.cuda.synchronize(self.device)
        self._data = CudaMemoryUsageData(
            starting_free_memory_bytes=self._starting_free_memory,
            starting_alloc_bytes=self._starting_alloc,
            max_alloc_bytes=torch.cuda.max_memory_allocated(self.device),
            new_alloc_bytes=torch.cuda.max_memory_allocated(self.device) - self._starting_alloc,
            device_memory_bytes=self._device_memory,
            _memory_stats=torch.cuda.memory_stats(self.device),
        )

    def get_usage_data(self) -> CudaMemoryUsageData:
        if self._data is None:
            msg = "CudaMemoryTracker was not properly exited"
            raise RuntimeError(msg)
        return self._data



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/entity__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.data import fco_metadata__client_pb2 as tecton__proto_dot_data_dot_fco__metadata____client__pb2
from tecton_proto.validation import validator__client_pb2 as tecton__proto_dot_validation_dot_validator____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&tecton_proto/data/entity__client.proto\x12\x11tecton_proto.data\x1a$tecton_proto/common/id__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a,tecton_proto/data/fco_metadata__client.proto\x1a/tecton_proto/validation/validator__client.proto\"\xd9\x03\n\x06\x45ntity\x12\x34\n\tentity_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x08\x65ntityId\x12(\n\x10join_keys_legacy\x18\x03 \x03(\tR\x0ejoinKeysLegacy\x12\x41\n\x0c\x66\x63o_metadata\x18\x07 \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12V\n\x0fvalidation_args\x18\t \x01(\x0b\x32-.tecton_proto.validation.EntityValidationArgsR\x0evalidationArgs\x12@\n\x07options\x18\n \x03(\x0b\x32&.tecton_proto.data.Entity.OptionsEntryR\x07options\x12\x38\n\tjoin_keys\x18\x0b \x03(\x0b\x32\x1b.tecton_proto.common.ColumnR\x08joinKeys\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01J\x04\x08\x02\x10\x03J\x04\x08\x04\x10\x05J\x04\x08\x05\x10\x06J\x04\x08\x06\x10\x07J\x04\x08\x08\x10\tB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.entity__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _ENTITY_OPTIONSENTRY._options = None
  _ENTITY_OPTIONSENTRY._serialized_options = b'8\001'
  _ENTITY._serialized_start=237
  _ENTITY._serialized_end=710
  _ENTITY_OPTIONSENTRY._serialized_start=622
  _ENTITY_OPTIONSENTRY._serialized_end=680
# @@protoc_insertion_point(module_scope)



import datetime
from collections import defaultdict
from typing import Dict
from typing import List
from typing import Optional

from tecton_core import aggregation_utils
from tecton_core import data_types as tecton_types
from tecton_core import errors
from tecton_core import query_consts
from tecton_core import schema as core_schema
from tecton_core import specs
from tecton_core.aggregation_utils import get_aggregation_function_result_type
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.specs import LifetimeWindowSpec
from tecton_core.specs import RelativeTimeWindowSpec
from tecton_core.specs import utils
from tecton_core.specs import window_spec_to_window_data_proto
from tecton_proto.args import feature_view__client_pb2 as feature_view__args_pb2
from tecton_proto.args import feature_view__client_pb2 as feature_view_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2


def _get_timestamp_field(feature_view_args: feature_view_pb2.FeatureViewArgs, view_schema: schema_pb2.Schema) -> str:
    timestamp_key = ""

    if feature_view_args.materialized_feature_view_args.HasField("timestamp_field"):
        timestamp_key = feature_view_args.materialized_feature_view_args.timestamp_field
    else:
        timestamp_fields = [
            column for column in view_schema.columns if column.offline_data_type == tecton_types.TimestampType().proto
        ]

        if len(timestamp_fields) != 1:
            msg = "The timestamp_field must be set on the Feature View or the feature view transformation output should contain only one and only one column of type Timestamp"
            raise errors.TectonValidationError(msg)
        timestamp_key = timestamp_fields[0].name

    view_schema_column_names = [column.name for column in view_schema.columns]
    if timestamp_key not in view_schema_column_names:
        msg = f"Timestamp key '{timestamp_key}' not found in view schema. View schema has columns: {', '.join(view_schema_column_names)}"
        raise errors.TectonValidationError(msg)
    return timestamp_key


def populate_schema_with_derived_fields(schema: schema_pb2.Schema) -> None:
    """Copies the behavior of populateSchemaWithDerivedFields in FeatureViewUtils.kt.

    Should only be applied to feature views w/ push sources or explicit schemas in NDD,
    which are expected to only have the offline_data_type field set.
    """
    for column in schema.columns:
        assert column.offline_data_type is not None
        offline_data_type = tecton_types.data_type_from_proto(column.offline_data_type)
        feature_server_data_type = core_schema.get_feature_server_data_type(offline_data_type)
        column.feature_server_data_type.CopyFrom(feature_server_data_type.proto)


def _get_aggregation_columns(
    view_schema_column_map: Dict[str, schema_pb2.Column],
    aggregation: feature_view_pb2.FeatureAggregation,
    is_continuous: bool,
) -> List[schema_pb2.Column]:
    input_column = view_schema_column_map[aggregation.column]
    prefixes = aggregation_utils.get_materialization_aggregation_column_prefixes(
        aggregation_utils.get_aggregation_enum_from_string(aggregation.function.lower()),
        aggregation.function_params,
        is_continuous,
    )
    columns = []
    for prefix in prefixes:
        materialization_column_name = aggregation_utils.get_materialization_column_name(prefix, input_column.name)

        tecton_type = aggregation_utils.aggregation_prefix_to_tecton_type(
            prefix, tecton_types.data_type_from_proto(input_column.offline_data_type)
        )
        if tecton_type is None:
            tecton_type = tecton_types.data_type_from_proto(input_column.offline_data_type)

        column_proto = schema_pb2.Column()
        column_proto.CopyFrom(core_schema.column_from_tecton_data_type(tecton_type))
        column_proto.name = materialization_column_name
        columns.append(column_proto)

    return columns


def compute_features_schema_from_feature_definition(
    fd: FeatureDefinitionWrapper,
) -> schema_pb2.Schema:
    if not fd.is_temporal_aggregate:
        return fd.materialization_schema.proto

    materialization_schema_columns = []
    view_schema_column_map = {column.name: column for column in fd.view_schema.proto.columns}

    for join_key in fd.join_keys:
        materialization_schema_columns.append(view_schema_column_map[join_key])

    if fd.aggregation_secondary_key:
        materialization_schema_columns.append(view_schema_column_map[fd.aggregation_secondary_key])

    for aggregation in fd.fv_spec.aggregate_features:
        input_column = view_schema_column_map[aggregation.input_feature_name]
        input_type = tecton_types.data_type_from_proto(input_column.offline_data_type)
        tecton_type = get_aggregation_function_result_type(aggregation.function, input_type)

        column_proto = schema_pb2.Column()
        column_proto.CopyFrom(core_schema.column_from_tecton_data_type(tecton_type))
        column_proto.name = aggregation.output_feature_name
        materialization_schema_columns.append(column_proto)

    column_proto = schema_pb2.Column()
    column_proto.CopyFrom(core_schema.column_from_tecton_data_type(tecton_types.Int64Type()))
    column_proto.name = query_consts.anchor_time()
    materialization_schema_columns.append(column_proto)

    return schema_pb2.Schema(columns=materialization_schema_columns)


def compute_aggregate_materialization_schema_from_view_schema(
    view_schema: schema_pb2.Schema,
    feature_view_args: feature_view_pb2.FeatureViewArgs,
    use_anchor_time: bool,
) -> schema_pb2.Schema:
    is_aggregate = len(feature_view_args.materialized_feature_view_args.aggregations) > 0
    has_embeddings = len(feature_view_args.materialized_feature_view_args.embeddings) > 0
    if not is_aggregate and not has_embeddings:
        return view_schema

    materialization_schema_columns = []
    view_schema_column_map = {column.name: column for column in view_schema.columns}

    # Add join key columns from view schema to materialization schema.
    join_keys = specs.get_join_keys_from_feature_view_args(feature_view_args)
    for join_key in join_keys:
        if join_key not in view_schema_column_map:
            msg = f"Join key {join_key} not found in view schema. View schema has columns {','.join(view_schema_column_map.keys())}"
            raise errors.TectonValidationError(msg)
        materialization_schema_columns.append(view_schema_column_map[join_key])

    if is_aggregate:
        if feature_view_args.materialized_feature_view_args.aggregation_secondary_key:
            materialization_schema_columns.append(
                view_schema_column_map[feature_view_args.materialized_feature_view_args.aggregation_secondary_key]
            )

        # Add columns for aggregate features.
        added = []
        for aggregation in feature_view_args.materialized_feature_view_args.aggregations:
            if aggregation.column not in view_schema_column_map:
                msg = f"Column {aggregation.column} used for aggregations not found in view schema. View schema has columns {','.join(view_schema_column_map.keys())}"
                raise errors.TectonValidationError(msg)

            is_continuous = specs.get_is_continuous_from_feature_view_args(feature_view_args)

            columns = _get_aggregation_columns(view_schema_column_map, aggregation, is_continuous)

            for col in columns:
                if col.name not in added:
                    materialization_schema_columns.append(col)
                    added.append(col.name)

        # Add column for timestamp. For Spark, aggregate feature views use an anchor time column.
        if use_anchor_time:
            column_proto = schema_pb2.Column()
            column_proto.CopyFrom(core_schema.column_from_tecton_data_type(tecton_types.Int64Type()))
            column_proto.name = query_consts.anchor_time()
            materialization_schema_columns.append(column_proto)
        else:
            timestamp_key = _get_timestamp_field(feature_view_args, view_schema)
            materialization_schema_columns.append(view_schema_column_map[timestamp_key])
    elif has_embeddings:
        for attribute in feature_view_args.materialized_feature_view_args.attributes:
            if attribute.column not in view_schema_column_map:
                msg = f"Attribute {attribute.column} not found in view schema. View schema has columns {','.join(view_schema_column_map.keys())}"
                raise errors.TectonValidationError(msg)
            materialization_schema_columns.append(view_schema_column_map[attribute.column])

        timestamp_key = _get_timestamp_field(feature_view_args, view_schema)
        materialization_schema_columns.append(view_schema_column_map[timestamp_key])

        for f in feature_view_args.materialized_feature_view_args.embeddings:
            column_proto = schema_pb2.Column()
            column_proto.CopyFrom(
                core_schema.column_from_tecton_data_type(tecton_types.ArrayType(tecton_types.Float32Type()))
            )
            column_proto.name = f.name
            materialization_schema_columns.append(column_proto)

    return schema_pb2.Schema(columns=materialization_schema_columns)


def compute_batch_table_format(
    feature_view_args: feature_view_pb2.FeatureViewArgs,
    view_schema: schema_pb2.Schema,
) -> schema_pb2.OnlineBatchTableFormat:
    aggregations = feature_view_args.materialized_feature_view_args.aggregations
    if aggregations:
        is_continuous = (
            feature_view_args.materialized_feature_view_args.stream_processing_mode
            == feature_view_pb2.StreamProcessingMode.STREAM_PROCESSING_MODE_CONTINUOUS
        )
        return _compute_batch_table_schema_for_aggregate_feature_view(view_schema, aggregations, is_continuous)

    join_keys = specs.get_join_keys_from_feature_view_args(feature_view_args)
    ttl = utils.get_duration_field_or_none(feature_view_args.materialized_feature_view_args, "serving_ttl")
    timestamp_field = specs.resolve_timestamp_field(feature_view_args, view_schema)
    return _compute_batch_table_schema_for_temporal_feature_view(view_schema, join_keys, timestamp_field, ttl)


def _compute_batch_table_schema_for_temporal_feature_view(
    view_schema: schema_pb2.Schema,
    join_keys: List[str],
    timestamp_key: str,
    serving_ttl: Optional[datetime.timedelta],
) -> schema_pb2.OnlineBatchTableFormat:
    if serving_ttl is None:
        window = LifetimeWindowSpec()
    else:
        window = RelativeTimeWindowSpec(-1 * serving_ttl, datetime.timedelta())
    output_columns = (col for col in view_schema.columns if col.name not in join_keys and col.name != timestamp_key)
    output_schema = schema_pb2.Schema(columns=sorted(output_columns, key=lambda col: col.name))
    part = schema_pb2.OnlineBatchTablePart(
        window_index=0, time_window=window_spec_to_window_data_proto(window), schema=output_schema
    )
    return schema_pb2.OnlineBatchTableFormat(online_batch_table_parts=[part])


def _compute_batch_table_schema_for_aggregate_feature_view(
    view_schema: schema_pb2.Schema,
    aggregations: List[feature_view__args_pb2.FeatureAggregation],
    is_continuous: bool,
) -> schema_pb2.OnlineBatchTableFormat:
    view_schema_column_map = {column.name: column for column in view_schema.columns}

    grouped_aggs = defaultdict(list)

    # Group by
    for aggregation in aggregations:
        if aggregation.HasField("time_window"):
            key = RelativeTimeWindowSpec.from_args_proto(aggregation.time_window)
        elif aggregation.HasField("lifetime_window"):
            key = LifetimeWindowSpec()
        else:
            msg = f"Unexpected time window type in aggregation args proto: {aggregation}"
            raise ValueError(msg)

        grouped_aggs[key].append(aggregation)

    # Sort by two keys
    sorted_grouped_aggs = sorted(grouped_aggs.items(), key=lambda item: item[0].to_sort_tuple())

    parts = []

    for index, (window, sub_aggs) in enumerate(sorted_grouped_aggs):
        added = set()
        columns = []
        for aggregation in sub_aggs:
            aggregation_columns = _get_aggregation_columns(
                view_schema_column_map, aggregation, is_continuous=is_continuous
            )
            for col in aggregation_columns:
                if col.name not in added:
                    columns.append(col)
                    added.add(col.name)

        output_schema = schema_pb2.Schema(columns=sorted(columns, key=lambda col: col.name))
        # We are NOT computing the OnlineBatchTablePartTiles here since they are not used during local development.
        part = schema_pb2.OnlineBatchTablePart(
            window_index=index, time_window=window_spec_to_window_data_proto(window), schema=output_schema
        )
        parts.append(part)

    return schema_pb2.OnlineBatchTableFormat(online_batch_table_parts=parts)



import contextlib
import logging
import traceback
import typing
from datetime import datetime
from typing import Optional

from tecton_core.query.executor_utils import QueryTreeMonitor
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_proto.materialization.job_metadata__client_pb2 import JobMetadata
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage


logger = logging.getLogger(__name__)

ProgressLogger = typing.Callable[[float], None]
SQLParam = typing.Optional[str]


class MonitoringContextProvider(typing.Protocol):
    def __call__(self, p: SQLParam = None) -> typing.ContextManager[ProgressLogger]: ...


class JobStatusClient(QueryTreeMonitor):
    def __init__(self, metadata_client: JobMetadataClient):
        self._metadata_client = metadata_client
        self._started = False
        self._query_count = 1
        self._query_index = 0

    def set_query_index(self, index: int, count: int):
        if count != self._query_count and self._started:
            msg = "Can't change query count after stages have been created"
            raise ValueError(msg)
        self._query_index = index
        self._query_count = count

    def create_stage(self, stage_type: TectonManagedStage.StageType, description: str) -> int:
        """
        Returns created stage index
        """
        self._started = True

        def _update(job_metadata: JobMetadata) -> Optional[JobMetadata]:
            if any(
                s.stage_type == stage_type and s.description == description
                for s in job_metadata.tecton_managed_info.stages
            ):
                return None
            new_proto = JobMetadata()
            new_proto.CopyFrom(job_metadata)

            new_stage = TectonManagedStage(
                description=description,
                stage_type=stage_type,
                state=TectonManagedStage.State.PENDING,
            )
            new_proto.tecton_managed_info.stages.append(new_stage)
            return new_proto

        metadata = self._metadata_client.update(_update)
        for i, stage in enumerate(metadata.tecton_managed_info.stages):
            if stage.stage_type == stage_type and stage.description == description:
                return i
        msg = f"Stage {stage_type} does not exist"
        raise ValueError(msg)

    def set_query(self, stage_idx, sql: str):
        def _update(job_metadata: JobMetadata) -> JobMetadata:
            new_proto = JobMetadata()
            new_proto.CopyFrom(job_metadata)

            stage = new_proto.tecton_managed_info.stages[stage_idx]
            if not stage.compiled_sql_query:
                stage.compiled_sql_query = sql

            return new_proto

        self._metadata_client.update(_update)

    def update_progress(self, stage_idx: int, progress: float):
        job_progress = (float(self._query_index) / float(self._query_count)) + progress * (
            1.0 / float(self._query_count)
        )

        def _update(job_metadata: JobMetadata) -> JobMetadata:
            new_proto = JobMetadata()
            new_proto.CopyFrom(job_metadata)

            stage = new_proto.tecton_managed_info.stages[stage_idx]

            if stage.state == TectonManagedStage.PENDING:
                stage.state = TectonManagedStage.State.RUNNING
                stage.start_time.GetCurrentTime()

            stage.progress = job_progress
            stage.duration.FromSeconds(int((datetime.now() - stage.start_time.ToDatetime()).total_seconds()))

            return new_proto

        self._metadata_client.update(_update)

    def set_completed(self, stage_idx: int):
        if self._query_index < (self._query_count - 1):
            return

        def _update(job_metadata: JobMetadata) -> JobMetadata:
            new_proto = JobMetadata()
            new_proto.CopyFrom(job_metadata)

            stage = new_proto.tecton_managed_info.stages[stage_idx]
            stage.state = TectonManagedStage.State.SUCCESS

            return new_proto

        self._metadata_client.update(_update)

    def set_failed(self, stage_idx: int, user_error: bool):
        def _update(job_metadata: JobMetadata) -> JobMetadata:
            new_proto = JobMetadata()
            new_proto.CopyFrom(job_metadata)

            stage = new_proto.tecton_managed_info.stages[stage_idx]
            stage.error_type = (
                TectonManagedStage.ErrorType.USER_ERROR if user_error else TectonManagedStage.ErrorType.UNEXPECTED_ERROR
            )
            stage.error_detail = traceback.format_exc()
            stage.state = TectonManagedStage.State.ERROR

            return new_proto

        self._metadata_client.update(_update)

    def set_current_stage_failed(self, error_type: TectonManagedStage.ErrorType):
        def _update(job_metadata: JobMetadata) -> JobMetadata:
            new_proto = JobMetadata()
            new_proto.CopyFrom(job_metadata)

            if any(s.state == TectonManagedStage.State.ERROR for s in new_proto.tecton_managed_info.stages):
                return new_proto

            current_stage = None
            for stage in new_proto.tecton_managed_info.stages:
                # Select first RUNNING or PENDING stage
                # Or if all stages are complete - just the last one
                current_stage = stage
                if stage.state in (TectonManagedStage.State.RUNNING, TectonManagedStage.State.PENDING):
                    break

            if not current_stage:
                # if there are no stages - we will create a dummy one
                current_stage = TectonManagedStage(
                    description="Setting up materialization job",
                    state=TectonManagedStage.State.ERROR,
                )
                new_proto.tecton_managed_info.stages.append(current_stage)

            current_stage.error_type = error_type
            current_stage.error_detail = traceback.format_exc()
            current_stage.state = TectonManagedStage.State.ERROR

            return new_proto

        self._metadata_client.update(_update)

    def create_stage_monitor(
        self,
        stage_type: TectonManagedStage.StageType,
        description: str,
    ) -> MonitoringContextProvider:
        stage_idx = self.create_stage(stage_type, description)

        @contextlib.contextmanager
        def monitor(sql: Optional[str] = None):
            if sql:
                self.set_query(stage_idx, sql)

            self.update_progress(stage_idx, 0)

            try:
                yield lambda p: self.update_progress(stage_idx, p)
            except Exception:
                self.set_failed(stage_idx, user_error=False)
                raise
            else:
                self.update_progress(stage_idx, 1)
                self.set_completed(stage_idx)

        return monitor

    def set_overall_state(self, state: TectonManagedStage.State) -> None:
        def _update(job_metadata: JobMetadata) -> JobMetadata:
            new_proto = JobMetadata()
            new_proto.CopyFrom(job_metadata)

            new_proto.tecton_managed_info.state = state

            return new_proto

        self._metadata_client.update(_update)
